# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import abc
import sys
from typing import Type

import numpy as np
from scipy.special import erf

from qat.utils.pydantic import AllowExtraFieldsModel

UPCONVERT_SIGN = 1.0
# use slightly below the maxmiumum allowed float to avoid flaky overflow errors
MAX_COSH_ARG = np.arccosh(0.99 * sys.float_info.max)


class ComplexFunction(AllowExtraFieldsModel, abc.ABC):
    """Function object used to represent Complex 1D functions."""

    _dtype: Type = np.complex128
    dt: float = 0.5e-9

    @abc.abstractmethod
    def eval(self, x: np.ndarray) -> np.ndarray:
        """
        Function evaluated in domain described by x.
        """
        pass

    @abc.abstractmethod
    def derivative(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        First order derivative.
        """
        pass

    def __call__(self, x: np.ndarray) -> np.ndarray:
        return self.eval(x)


def _validate_input_array(func):
    """
    Wrapper method to validate the input array.
    """

    def validator(*args, **kwargs):
        max_size = 1e6
        # Check every argument except for any `self`.
        for arg in [val for val in args if not isinstance(val, ComplexFunction)]:
            if not isinstance(arg, np.ndarray):
                raise TypeError(f"Function given {type(arg)}, expecting numpy.ndarray.")
            if arg.size > max_size:
                raise RuntimeError(
                    f"Function given {arg.size} element inputs "
                    f"exceeding maximum {max_size}."
                )
        return func(*args, **kwargs)

    return validator


class NumericFunction(ComplexFunction):
    """
    Base class for functions applying an numerical first derivative.
    """

    @_validate_input_array
    def derivative(self, x: np.ndarray, y: np.ndarray) -> np.ndarray:
        """
        For a custom wave-pulse or pulse without analytic derivative compute it
        numerically.
        """
        if len(x) < 2:
            return np.zeros_like(y)
        else:
            amplitude_differential = np.diff(y) / np.diff(x)
            # np.diff reduces the array by 1 item when calculating the differential,
            # so we approximate what the last point should look like
            return np.append(
                amplitude_differential,
                2 * amplitude_differential[-1] - amplitude_differential[-2],
            )


class SquareFunction(ComplexFunction):
    """
    Square function of fixed amplitude.
    """

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        return np.ones(shape=x.shape, dtype=self._dtype)

    @_validate_input_array
    def derivative(self, x: np.ndarray, _=None) -> np.ndarray:
        return np.zeros(shape=x.shape, dtype=self._dtype)


class GaussianFunction(ComplexFunction):
    """
    Gaussian function.
    """

    width: float
    rise: float

    @property
    def k(self):
        return self.width * self.rise

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        return np.exp(-((x / self.k) ** 2), dtype=self._dtype)

    @_validate_input_array
    def derivative(self, x: np.ndarray, _=None) -> np.ndarray:
        c = -2.0 * x / self.k**2
        return c * self.eval(x)


class GaussianZeroEdgeFunction(ComplexFunction):
    """
    A Gaussian pulse that can be normalized to be zero at the edges.
    """

    std_dev: float
    width: float
    zero_at_edges: bool

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        gauss = np.exp(-0.5 * (x / self.std_dev) ** 2)
        if self.zero_at_edges:
            zae_chunk = self.zero_at_edges * (
                np.exp(-0.5 * ((self.width / 2) / self.std_dev) ** 2)
            )
            gauss = (gauss - zae_chunk) / (1 - zae_chunk)
        return gauss

    def derivative(self, x: np.ndarray, _=None):
        raise NotImplementedError(
            f"`{self.__class__.__name__}` does not have an implementation for `derivative`."
        )


class GaussianSquareFunction(NumericFunction):
    """
    A square pulse with a Gaussian rise and fall at the edges.
    """

    square_width: float
    std_dev: float
    zero_at_edges: float

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        y = np.ones(shape=x.shape, dtype=self._dtype)
        x_rise = x[x < -self.square_width / 2] + (self.square_width / 2)
        x_fall = x[x > self.square_width / 2] - (self.square_width / 2)
        y[x < -self.square_width / 2] = np.exp(-0.5 * (x_rise / self.std_dev) ** 2)
        y[x > self.square_width / 2] = np.exp(-0.5 * (x_fall / self.std_dev) ** 2)
        if self.zero_at_edges:
            y = (y - y[0]) / (1 - y[0])
        return y


class DragGaussianFunction(ComplexFunction):
    """
    Drag Gaussian, tighter on one side and long tail on the other.
    """

    std_dev: float
    beta: float
    zero_at_edges: bool

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        zae_chunk = -self.zero_at_edges * (
            np.exp(-0.5 * (self.std_dev / self.std_dev) ** 2)
        )
        beta_chunk = 1 - 1j * self.beta * x / self.std_dev**2
        coef = beta_chunk / (1 - zae_chunk)
        gauss = np.exp(-0.5 * (x / (2 * self.std_dev)) ** 2)
        return coef * (gauss - zae_chunk)

    def derivative(self, x: np.ndarray, _=None):
        raise NotImplementedError(
            f"`{self.__class__.__name__}` does not have an implementation for `derivative`."
        )


class SechFunction(ComplexFunction):
    """
    Implements a sech pulse defined by sech(x / width). Note that it is not normalized to be
    zero at the edges.
    """

    std_dev: float

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        # Having a narrow width can cause overflows in numpy
        # Restricting the argument such that cosh is within the max float range
        # will overcome this, and has a neglibable effect (as sech(x) outside this
        # range is practically zero).
        x = np.clip(x.real / self.std_dev, -MAX_COSH_ARG, MAX_COSH_ARG)
        return 1 / np.cosh(x)

    def derivative(self, x: np.ndarray, _=None):
        raise NotImplementedError(
            f"`{self.__class__.__name__}` does not have an implementation for `derivative`."
        )


class Sin(ComplexFunction):
    frequency: float
    internal_phase: float

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        return np.sin(2 * np.pi * x * self.frequency + self.internal_phase)

    def derivative(self, x: np.ndarray, _=None):
        raise NotImplementedError(
            f"`{self.__class__.__name__}` does not have an implementation for `derivative`."
        )


class Cos(ComplexFunction):
    frequency: float
    internal_phase: float

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        return np.cos(2 * np.pi * x * self.frequency + self.internal_phase)

    def derivative(self, x: np.ndarray, _=None):
        raise NotImplementedError(
            f"`{self.__class__.__name__}` does not have an implementation for `derivative`."
        )


class RoundedSquareFunction(ComplexFunction):
    """
    Rounded square.
           ___
         /    \
     ___|      |___
    """

    width: float
    rise: float
    std_dev: float

    def step(self, val):
        return (erf(val) + 1) / 2

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        x1 = (self.width - self.std_dev) / 2
        x2 = (self.width + self.std_dev) / 2
        rescaling = (
            self.step((self.width / 2 - x1) / self.rise)
            + self.step(-(self.width / 2 - x2) / self.rise)
            - 1
        )
        return rescaling * (
            self.step((x - x1) / self.rise) + self.step(-(x - x2) / self.rise) - 1
        )

    def derivative(self, x: np.ndarray, _=None):
        raise NotImplementedError(
            f"`{self.__class__.__name__}` does not have an implementation for `derivative`."
        )


class BlackmanFunction(ComplexFunction):
    width: float

    a0: float = 7938.0 / 18608.0
    a1: float = 9240.0 / 18608.0
    a2: float = 1430.0 / 18608.0

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        nN = 2.0 * np.pi * (x / self.width + 0.5)
        return self.a0 - self.a1 * np.cos(nN) + self.a2 * np.cos(2 * nN)

    @_validate_input_array
    def derivative(self, x: np.ndarray, _=None) -> np.ndarray:
        nN = 2.0 * np.pi * (x / self.width + 0.5)
        c = 2.0 * np.pi / self.width
        return self.a1 * c * np.sin(nN) - self.a2 * 2 * c * np.sin(2 * nN)


class SoftSquareFunction(NumericFunction):
    width: float
    rise: float

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        pulse = 0.5 * (
            np.tanh((x.real + (self.width - self.rise) / 2.0) / self.rise)
            - np.tanh((x.real - (self.width - self.rise) / 2.0) / self.rise)
        )
        return pulse.astype(self._dtype)


class SofterSquareFunction(NumericFunction):
    std_dev: float
    rise: float

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        pulse = np.tanh((x.real + self.std_dev / 2.0 - self.rise) / self.rise) - np.tanh(
            (x.real - self.std_dev / 2.0 + self.rise) / self.rise
        )
        if pulse.any():
            pulse -= np.min(pulse)
            pulse /= np.max(pulse)
        return pulse.astype(self._dtype)


class ExtraSoftSquareFunction(NumericFunction):
    std_dev: float
    rise: float

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        pulse = np.tanh(
            (x.real + self.std_dev / 2.0 - 2.0 * self.rise) / self.rise
        ) - np.tanh((x.real - self.std_dev / 2.0 + 2.0 * self.rise) / self.rise)
        if pulse.any():
            pulse -= np.min(pulse)
            pulse /= np.max(pulse)
        return pulse.astype(self._dtype)


class SofterGaussianFunction(NumericFunction):
    width: float
    rise: float

    @_validate_input_array
    def eval(self, x: np.ndarray) -> np.ndarray:
        pulse = GaussianFunction(rise=self.rise, width=self.width).eval(x)
        if pulse.any():
            pulse -= min(pulse)
            pulse /= max(pulse)
        return np.array(pulse, dtype=self._dtype)


class SetupHoldFunction(NumericFunction):
    width: float
    rise: float
    amp_setup: float
    amp: float

    @_validate_input_array
    def eval(self, x):
        arr = np.ones(shape=x.shape, dtype=self._dtype)
        dt = x[1] - x[0]
        high_samples = int(self.rise / dt + 0.5)
        arr[0:high_samples] *= self.amp_setup / self.amp
        return arr



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

__all__ = [
    "EchoModelLoader",
    "FileModelLoader",
    "QiskitModelLoader",
    "RTCSModelLoader",
    "LucyModelLoader",
    "QbloxDummyModelLoader",
    "QbloxFileModelLoader",
]

from .dummy import QbloxDummyModelLoader
from .echo import EchoModelLoader
from .file import FileModelLoader, QbloxFileModelLoader
from .lucy import LucyModelLoader
from .qiskit import QiskitModelLoader
from .rtcs import RTCSModelLoader



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.core.config.descriptions import (
    CompilePipelineDescription,
    EngineDescription,
    ExecutePipelineDescription,
    HardwareLoaderDescription,
    PipelineClassDescription,
    PipelineFactoryDescription,
    PipelineInstanceDescription,
    UpdateablePipelineDescription,
)
from qat.engines import NativeEngine
from qat.engines.model import requires_hardware_model
from qat.model.hardware_model import PhysicalHardwareModel
from qat.model.loaders.base import BaseModelLoader
from qat.model.loaders.cache import CacheAccessLoader
from qat.model.loaders.update import ModelUpdateChecker
from qat.pipelines.base import AbstractPipeline
from qat.pipelines.cache import CompilePipelineCache, ExecutePipelineCache
from qat.pipelines.pipeline import CompilePipeline, ExecutePipeline, Pipeline
from qat.pipelines.updateable import UpdateablePipeline
from qat.purr.compiler.hardware_models import QuantumHardwareModel


class HardwareLoaders:
    def __init__(self, hardware_loaders: dict[str, BaseModelLoader] = {}):
        self._loaders = dict(**hardware_loaders)
        self._loaded_models = {}

    def __getitem__(self, model_name: str) -> QuantumHardwareModel | PhysicalHardwareModel:
        """Allows indexing to get a hardware model by name."""
        return self.load(model_name, allow_cache=True)

    def get_loader(self, loader_name: str, default=None) -> BaseModelLoader:
        """Returns a hardware model loader by name."""
        loader = self._loaders.get(loader_name)
        if loader is None:
            return default
        return loader

    def load(self, loader_name: str, default=None, allow_cache=True):
        """Loads a hardware model, using the internal cache unless `allow_cache=False`."""
        model = self._loaded_models.get(loader_name) if allow_cache else None
        if model is None:
            loader = self._loaders.get(loader_name)
            if loader is None:
                return default
            model = loader.load()
            self._loaded_models[loader_name] = model

        return model

    def clear_cache(self):
        self._loaded_models = {}

    def reload_model(
        self, loader_name: str
    ) -> QuantumHardwareModel | PhysicalHardwareModel:
        """Reloads a hardware model from its loader, updating the model in the cache."""
        if loader_name not in self._loaders:
            raise Exception(f"Hardware Model Loader {loader_name} not found")
        return self.load(loader_name, allow_cache=False)

    def reload_all_models(self):
        """Reloads all hardware models from their respective loaders, updating the cache."""
        for loader_name in self._loaders:
            self.reload_model(loader_name)

    @property
    def models_up_to_date(self) -> bool:
        """Checks if all models are up-to-date, for loaders that implement
        :class:`ModelUpdateChecker`.
        """
        for key, loader in self._loaders.items():
            model = self._loaded_models.get(key, None)
            if model is None:
                return False
            if isinstance(loader, ModelUpdateChecker) and not loader.is_up_to_date(model):
                return False
        return True

    @classmethod
    def from_descriptions(
        cls,
        hardware_loader_descriptions: list[HardwareLoaderDescription],
    ):
        return cls({hld.name: hld.construct() for hld in hardware_loader_descriptions})


class EngineSet:
    """Stores a set of engines, allowing identification by name.

    Also stores corresponding loaders which can be used to update the models within engines
    for engines that have a :class:`RequiresHardwareModelMixin` mixin.
    """

    def __init__(
        self,
        engines: dict[str, NativeEngine] = {},
        loaders: dict[str, BaseModelLoader] = {},
    ):
        self._engines: dict[str, NativeEngine] = dict(**engines)
        self._loaders: dict[str, BaseModelLoader] = dict(**loaders)

    def __getitem__(self, name: str) -> NativeEngine:
        """Allows indexing to get an engine by name."""
        return self.get(name)

    def get(self, name: str, default=None) -> NativeEngine:
        """Returns an engine by name, or a default if not found."""
        return self._engines.get(name, default)

    def reload_model(self, name: str):
        """Reloads the model for an engine with the given identifier."""
        if name not in self._engines:
            raise ValueError(f"Engine {name} not found")

        engine = self._engines[name]
        # TODO: Change to just check against RequiresHardwareModelMixin (COMPILER-662)
        if not requires_hardware_model(engine):
            return

        loader = self._loaders.get(name, None)
        if loader is None:
            return

        model = loader.load()
        engine.model = model

    def reload_all_models(self):
        """Reloads all models for engines that require a hardware model."""
        for name in self._engines:
            self.reload_model(name)

    @classmethod
    def from_descriptions(
        cls,
        engine_descriptions: list[EngineDescription],
        available_hardware: HardwareLoaders,
    ) -> "EngineSet":
        """Creates an :class:`EngineSet` from a list of engine descriptions."""

        engines = {}
        loaders = {}
        for desc in engine_descriptions:
            if desc.hardware_loader:
                loader = CacheAccessLoader(available_hardware, desc.hardware_loader)
                loaders[desc.name] = loader
                engines[desc.name] = desc.construct(model=loader.load())
            else:
                engines[desc.name] = desc.construct(model=None)
        return cls(engines, loaders)


class PipelineSet:
    def __init__(self, pipelines: list[AbstractPipeline] = []):
        self._pipelines = {}
        self._default_pipeline = None

        for P in pipelines:
            self.add(P)

    @classmethod
    def from_descriptions(
        cls,
        pipeline_descriptions: list[
            PipelineInstanceDescription
            | PipelineFactoryDescription
            | UpdateablePipelineDescription
            | ExecutePipelineDescription
            | CompilePipelineDescription
            | PipelineClassDescription
        ],
        available_hardware: HardwareLoaders,
        available_engines: EngineSet,
    ):
        pipeline_descriptions = (
            [] if pipeline_descriptions is None else pipeline_descriptions
        )
        default = next(
            (Pdesc.name for Pdesc in pipeline_descriptions if Pdesc.default), None
        )

        pipes = []
        for Pdesc in pipeline_descriptions:
            attrs = dict()
            if hasattr(Pdesc, "hardware_loader"):
                attrs["loader"] = CacheAccessLoader(
                    available_hardware, Pdesc.hardware_loader
                )
            if hasattr(Pdesc, "engine"):
                attrs["engine"] = available_engines[Pdesc.engine]
            P = Pdesc.construct(**attrs)
            pipes.append(P)

        pipelinesset = cls(pipes)
        if default:
            pipelinesset.set_default(default)
        else:
            pipelinesset._default_pipeline = None

        return pipelinesset

    def set_default(self, pipeline: AbstractPipeline | str):
        if isinstance(pipeline, AbstractPipeline):
            name = pipeline.name
            if name not in self._pipelines:
                raise Exception("Add pipeline using add before setting default")
        else:
            name = pipeline
            if name not in self._pipelines:
                raise Exception(f"Cannot set default pipeline to unknown pipeline {name}")

        self._default_pipeline = name

    @property
    def default(self) -> str:
        """Returns the name of the current default pipeline"""
        return self._default_pipeline

    def add(self, pipeline: AbstractPipeline, default: bool = False):
        """Adds a pipeline for subsequent use for compilation and execution

        :param pipeline: A pipeline instance to add, indexed by pipeline.name
        :param default: Set the added pipeline as the default, defaults to False
        """
        self._pipelines[pipeline.name] = pipeline
        if default:
            self.set_default(pipeline)

    def remove(self, pipeline: AbstractPipeline | str):
        """Remove a pipeline

        :param pipeline: The name of a pipeline or a pipeline instance to remove
        """
        name = pipeline.name if isinstance(pipeline, AbstractPipeline) else pipeline

        if name not in self._pipelines:
            raise ValueError(f"Pipeline {name} not found")

        if isinstance(pipeline, AbstractPipeline):
            if pipeline is not self._pipelines[name]:
                raise ValueError(
                    f"Pipeline {name} is not the same as stored pipeline with the same name"
                )

        if self._default_pipeline == name:
            self._default_pipeline = None

        del self._pipelines[name]

    def get(self, pipeline: AbstractPipeline | str) -> AbstractPipeline:
        """Gets a stored pipeline by name (str) or passes through a pipeline instance

        :param pipeline: A pipeline instance or the string name of a stored pipeline
        """

        if isinstance(pipeline, str):
            if pipeline == "default":
                if self._default_pipeline is None:
                    raise Exception("No Default Pipeline Set")
                pipeline = self._default_pipeline
            elif pipeline not in self._pipelines:
                raise Exception(f"Pipeline {pipeline} not found")

            pipeline = self._pipelines[pipeline]

        return pipeline

    def reload_model(self, pipeline: str):
        """Refreshes a pipeline by updating the models from its cache.

        :param pipeline: The name of the pipeline to refresh.
        """

        pipeline = self._pipelines[pipeline]
        if not (isinstance(pipeline, UpdateablePipeline) and pipeline.has_loader):
            raise Exception(
                f"The pipeline {pipeline} is not an Updateable pipelines equipped with a "
                "loader."
            )
        pipeline.update(reload_model=True)

    def reload_all_models(self):
        """Refreshes all :class:`UpdateablePipeline` instances with hardware model loaders
        by updating the models from their caches."""

        for pipeline in self._pipelines.values():
            if isinstance(pipeline, UpdateablePipeline) and pipeline.has_loader:
                pipeline.update(reload_model=True)

    def list(self) -> list[str]:
        """Returns a list of available pipeline names"""
        return list(self._pipelines.keys())

    def __repr__(self):
        return str(self.list())

    def __contains__(self, pipeline: str) -> bool:
        """Checks if a pipeline is in the set by name."""

        if pipeline == "default":
            return self._default_pipeline is not None
        return pipeline in self._pipelines


class PipelineManager:
    """A manager for pipelines, storing compilation and execution pipelines.

    Currently, pipelines are catergorised by their type (or the type produced by
    factories):

    * :class:`CompilePipeline` - pipelines for compilation
    * :class:`ExecutePipeline` - pipelines for execution
    * :class:`Pipeline` - pipelines that handle both compilation and execution.

    As such, when making calls to pipelines by name, we need to handle the logic for
    retreiving the correct pipeline given the task.
    """

    def __init__(
        self,
        compile_pipelines: PipelineSet,
        execute_pipelines: PipelineSet,
        full_pipelines: PipelineSet,
    ):
        self._full_pipelines = full_pipelines
        self.compile_pipelines = compile_pipelines
        self.execute_pipelines = execute_pipelines

        for pipeline in self._full_pipelines.list():
            P = self._full_pipelines.get(pipeline)
            self._add_full_pipeline_to_compile_and_execute(
                P, default=P.name == self._full_pipelines.default
            )

    def get_compile_pipeline(self, pipeline: str | AbstractPipeline) -> AbstractPipeline:
        """Returns a compile pipeline by name or instance."""
        return self.compile_pipelines.get(pipeline)

    def get_execute_pipeline(self, pipeline: str | AbstractPipeline) -> AbstractPipeline:
        """Returns an execute pipeline by name or instance."""
        return self.execute_pipelines.get(pipeline)

    @property
    def list_compile_pipelines(self) -> list[str]:
        """Returns a list of compile pipeline names."""
        return self.compile_pipelines.list()

    @property
    def list_execute_pipelines(self) -> list[str]:
        """Returns a list of execute pipeline names."""
        return self.execute_pipelines.list()

    @property
    def default_compile_pipeline(self) -> str:
        """Returns the name of the default compile pipeline."""
        return self.compile_pipelines.default

    @property
    def default_execute_pipeline(self) -> str:
        """Returns the name of the default execute pipeline."""
        return self.execute_pipelines.default

    def add(self, pipeline: AbstractPipeline, default: bool = False):
        """Adds a pipeline to the manager.

        :param pipeline: The pipeline to add.
        :param default: If True, sets this pipeline as the default.
        """
        name = pipeline.name

        if pipeline.is_subtype_of(Pipeline):
            if name in self.list_compile_pipelines or name in self.list_execute_pipelines:
                raise ValueError(
                    f"Pipeline {name} already exists in the manager as a compile or "
                    "execute pipeline."
                )
            self._full_pipelines.add(pipeline)
            self._add_full_pipeline_to_compile_and_execute(pipeline, default)

        elif pipeline.is_subtype_of(CompilePipeline):
            if name in self.list_compile_pipelines:
                raise ValueError(
                    f"Pipeline {name} already exists in the manager as a compile pipeline."
                )
            self.compile_pipelines.add(pipeline, default)

        elif pipeline.is_subtype_of(ExecutePipeline):
            if name in self.list_execute_pipelines:
                raise ValueError(
                    f"Pipeline {name} already exists in the manager as an execute pipeline."
                )
            self.execute_pipelines.add(pipeline, default)

        else:
            raise ValueError(
                f"Pipeline {pipeline.name} is not a CompilePipeline or ExecutePipeline."
            )

    def remove(
        self,
        pipeline: AbstractPipeline | str,
        compile: bool | None = None,
        execute: bool | None = None,
    ):
        """Removes a pipeline from the manager.

        :param pipeline: The pipeline to remove, either by name or instance.
        :param compile: Determines if the pipeline should be removed from compile pipelines.
            If None, the manager will decide given the pipeline type.
        :param execute: Determines if the pipeline should be removed from execute pipelines.
            If None, the manager will decide given the pipeline type.
        """
        pipeline = pipeline.name if isinstance(pipeline, AbstractPipeline) else pipeline

        if compile is False and execute is False:
            return

        full = pipeline in self._full_pipelines
        if full and (compile is False or execute is False):
            raise ValueError(
                f"Cannot remove the full pipeline {pipeline} from just compile or execute "
                "pipelines. You must remove it from both."
            )

        compile = compile if compile is not None else pipeline in self.compile_pipelines
        if compile:
            self.compile_pipelines.remove(pipeline)

        execute = execute if execute is not None else pipeline in self.execute_pipelines
        if execute:
            self.execute_pipelines.remove(pipeline)

        if full:
            self._full_pipelines.remove(pipeline)

    def set_default(
        self,
        pipeline: AbstractPipeline | str,
        compile: bool | None = None,
        execute: bool | None = None,
    ):
        """Sets the default pipeline for compilation or execution.

        If the pipeline is a full pipeline, then it must be set as the default for both
        compilation and execution. Else, if there exists a compile or execute pipeline with
        the same name, it can be specified for which the default should be set for. By
        default, it will be both.
        """
        if compile is False and execute is False:
            return

        if isinstance(pipeline, AbstractPipeline):
            pipeline = pipeline.name

        if pipeline in self._full_pipelines:
            if compile is False or execute is False:
                raise ValueError(
                    f"Cannot set the full pipeline {pipeline} as default for just "
                    "compile or execute pipelines. It must be set for both."
                )
            compile = True if compile is None else compile
            execute = True if execute is None else execute

        if pipeline in self.compile_pipelines:
            compile = True if compile is None else compile
        elif compile is True:
            raise ValueError(
                f"Cannot set the pipeline {pipeline} as default for compile pipelines, as "
                "it cannot be found."
            )

        if pipeline in self.execute_pipelines:
            execute = True if execute is None else execute
        elif execute is True:
            raise ValueError(
                f"Cannot set the pipeline {pipeline} as default for execute pipelines, as "
                "it cannot be found."
            )

        if compile:
            self.compile_pipelines.set_default(pipeline)
        if execute:
            self.execute_pipelines.set_default(pipeline)

    def reload_all_models(self):
        """Reloads all models in all pipelines that have a loader."""
        self._full_pipelines.reload_all_models()
        self.compile_pipelines.reload_all_models()
        self.execute_pipelines.reload_all_models()

    @classmethod
    def from_descriptions(
        cls,
        compile_pipelines: list[
            PipelineInstanceDescription
            | PipelineFactoryDescription
            | UpdateablePipelineDescription
            | CompilePipelineDescription
        ],
        execute_pipelines: list[
            PipelineInstanceDescription
            | PipelineFactoryDescription
            | UpdateablePipelineDescription
            | ExecutePipelineDescription
        ],
        full_pipelines: list[
            PipelineInstanceDescription
            | PipelineFactoryDescription
            | UpdateablePipelineDescription
            | PipelineClassDescription
        ],
        available_hardware: HardwareLoaders,
        available_engines: EngineSet,
    ) -> "PipelineManager":
        """Creates a PipelineManager from a list of pipeline descriptions."""
        compile_set = PipelineSet.from_descriptions(
            compile_pipelines, available_hardware, available_engines
        )
        execute_set = PipelineSet.from_descriptions(
            execute_pipelines, available_hardware, available_engines
        )
        full_set = PipelineSet.from_descriptions(
            full_pipelines, available_hardware, available_engines
        )

        return cls(compile_set, execute_set, full_set)

    def _add_full_pipeline_to_compile_and_execute(self, P: Pipeline, default: bool = False):
        """Adds a full pipeline to the manager, updating both compile and execute pipelines."""
        self.compile_pipelines.add(CompilePipelineCache(P), default)
        self.execute_pipelines.add(ExecutePipelineCache(P), default)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd


from qat.core.pass_base import PassManager
from qat.middleend.base import CustomMiddleend


class FallthroughMiddleend(CustomMiddleend):
    """
    A middle end that passes through an input :class:`InstructionBuilder` and does not alter it.
    """

    def __init__(self, model: None = None):
        super().__init__(model=None, pipeline=PassManager())



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from typing import Dict, List

import numpy as np
from matplotlib import pyplot as plt

from qat.purr.backends.qblox.acquisition import Acquisition
from qat.purr.backends.qblox.codegen import QbloxPackage
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


def plot_packages(packages: List[QbloxPackage]):
    if not packages:
        return

    max_length = max([len(pkg.timeline) for pkg in packages])
    if max_length <= 0:
        return

    # Padding short timelines with zeros
    for pkg in packages:
        length = len(pkg.timeline)
        if length < max_length:
            pkg.timeline = np.append(
                pkg.timeline, np.zeros(max_length - length, dtype=pkg.timeline.dtype)
            )

    t = np.linspace(0, max_length, max_length)
    fig, axes = plt.subplots(
        nrows=len(packages),
        ncols=1,
        sharex=False,
        sharey=False,
        squeeze=False,
        figsize=(10, 5),
    )
    fig.suptitle("Timeline plots")
    for i, pkg in enumerate(packages):
        axes[i][0].plot(t, pkg.timeline.real, label="I")
        axes[i][0].plot(t, pkg.timeline.imag, label="Q")
        axes[i][0].set_title(pkg.pulse_channel_id)
        axes[i][0].set_xlabel("Time (ns)")
        axes[i][0].set_ylabel("Digital offset")
        axes[i][0].autoscale()
        axes[i][0].legend()

    plt.tight_layout()
    plt.show()


def plot_playback(playback: Dict[str, List[Acquisition]]):
    if not playback:
        return

    for i, (pulse_channel_id, acquisitions) in enumerate(playback.items()):
        for acquisition in acquisitions:
            fig, axes = plt.subplots(
                nrows=3,
                ncols=1,
                sharex=False,
                sharey=False,
                squeeze=False,
                figsize=(10, 5),
            )
            fig.suptitle(f"Playback plots for {acquisition.name} on {pulse_channel_id}")

            scope_data = acquisition.acquisition.scope
            integ_data = acquisition.acquisition.bins.integration
            thrld_data = acquisition.acquisition.bins.threshold

            # Scope data
            axes[0, 0].plot(scope_data.path0.data, label="I")
            axes[0, 0].plot(scope_data.path1.data, label="Q")
            axes[0, 0].set_xlabel("Sample (ns)")
            axes[0, 0].set_ylabel("Value")
            axes[0, 0].autoscale()
            axes[0, 0].legend()
            axes[0, 0].title.set_text("Scope acquisition")

            # Integration data
            axes[1, 0].plot(integ_data.path0, label="I")
            axes[1, 0].plot(integ_data.path1, label="Q")
            axes[1, 0].set_xlabel("Iteration")
            axes[1, 0].set_ylabel("Value")
            axes[1, 0].autoscale()
            axes[1, 0].legend()
            axes[1, 0].title.set_text("Integrated acquisition")

            # Threshold data
            axes[2, 0].plot(thrld_data, label="I")
            axes[2, 0].set_xlabel("Iteration")
            axes[2, 0].set_ylabel("Value")
            axes[2, 0].autoscale()
            axes[2, 0].title.set_text("Thresholded acquisition")

        plt.tight_layout()
        plt.show()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
import numpy as np
import regex
from qblox_instruments import (
    Cluster,
    DummyBinnedAcquisitionData,
    DummyScopeAcquisitionData,
    Sequencer,
)

from qat.backend.qblox.config.constants import Constants
from qat.backend.qblox.execution import QbloxProgram
from qat.backend.qblox.ir import Sequence
from qat.engines.qblox.live import QbloxLeafInstrument
from qat.purr.compiler.devices import ChannelType
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class DummyQbloxInstrument(QbloxLeafInstrument):
    shot_pattern = regex.compile("jlt( +)R([0-9]+),([0-9]+),@(.*)\n")

    def __init__(self, id, name, address, dummy_config):
        super().__init__(id, name, address)
        self.dummy_config = dummy_config

    def _setup_dummy_scope_acq_data(self, sequencer: Sequencer, sequence: Sequence):
        shot_match = next(self.shot_pattern.finditer(sequence.program), None)
        avg_count = int(shot_match.group(3)) if shot_match else 1

        dummy_data = np.random.random(
            size=(Constants.MAX_SAMPLE_SIZE_SCOPE_ACQUISITIONS, 2)
        )
        dummy_data = [(iq[0], iq[1]) for iq in dummy_data]
        dummy_scope_acquisition_data = DummyScopeAcquisitionData(
            data=dummy_data, out_of_range=(False, False), avg_cnt=(avg_count, avg_count)
        )
        sequencer.set_dummy_scope_acquisition_data(data=dummy_scope_acquisition_data)

    def _setup_dummy_binned_acq_data(self, sequencer: Sequencer, sequence: Sequence):
        shot_match = next(self.shot_pattern.finditer(sequence.program), None)
        avg_count = int(shot_match.group(3)) if shot_match else 1

        for name, acquisition in sequence.acquisitions.items():
            dummy_binned_acquisition_data = [
                DummyBinnedAcquisitionData(
                    data=(np.random.random(), np.random.random()),
                    thres=np.random.choice(2),
                    avg_cnt=avg_count,
                )
            ] * acquisition["num_bins"]
            sequencer.set_dummy_binned_acquisition_data(
                acq_index_name=name,
                data=dummy_binned_acquisition_data,
            )

    def _delete_acquisitions(self, sequencer):
        sequencer.delete_dummy_scope_acquisition_data()
        sequencer.delete_dummy_binned_acquisition_data()

    def connect(self):
        if self._driver is None or not Cluster.is_valid(self._driver):
            self._driver: Cluster = Cluster(name=self.name, dummy_cfg=self.dummy_config)
            self._driver.reset()
            self._modules = {m: True for m in self._driver.get_connected_modules().values()}
            self.is_connected = True

        log.info(self._driver.get_system_status())

    def setup(self, program: QbloxProgram):
        super().setup(program)

        # Stage Scope and Acquisition data
        for pulse_channel_id, sequencer in self._id2seq.items():
            if ChannelType.macq.name in pulse_channel_id:
                package = program.packages[pulse_channel_id]
                self._setup_dummy_scope_acq_data(sequencer, package.sequence)
                self._setup_dummy_binned_acq_data(sequencer, package.sequence)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from dataclasses import dataclass, field

from pydantic import ConfigDict

from qat.backend.qblox.config.specification import ModuleConfig, SequencerConfig
from qat.backend.qblox.ir import Sequence
from qat.executables import AbstractProgram
from qat.utils.pydantic import ComplexNDArray


@dataclass
class QbloxPackage:
    pulse_channel_id: str | None = None
    physical_channel_id: str | None = None
    instrument_id: str | None = None
    seq_idx: int | None = None
    seq_config: SequencerConfig = field(default_factory=lambda: SequencerConfig())
    slot_idx: int | None = None
    mod_config: ModuleConfig = field(default_factory=lambda: ModuleConfig())
    sequence: Sequence | None = None
    timeline: ComplexNDArray | None = None


class QbloxProgram(AbstractProgram):
    model_config = ConfigDict(arbitrary_types_allowed=True)

    # COMPILER 828: change to pydantic objects (not dataclasses)
    packages: dict[str, QbloxPackage]

    @property
    def acquire_shapes(self) -> dict[str, tuple[int, ...]]:
        return {}



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd
from qat.pipelines.updateable import PipelineConfig as PipelineConfig

from .purr.waveform_v1 import (
    EchoPipeline as EchoPipeline,
    echo8 as echo8,
    echo16 as echo16,
    echo32 as echo32,
)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import uuid

import numpy as np
import regex
from qblox_instruments import (
    ClusterType,
    DummyBinnedAcquisitionData,
    DummyScopeAcquisitionData,
)
from qblox_instruments.qcodes_drivers.sequencer import Sequencer

from qat.purr.backends.echo import Connectivity
from qat.purr.backends.qblox.codegen import QbloxPackage
from qat.purr.backends.qblox.config import (
    AttConfig,
    ConnectionConfig,
    LoConfig,
    ModuleConfig,
    OffsetConfig,
    QbloxConfig,
    SequencerConfig,
)
from qat.purr.backends.qblox.constants import Constants
from qat.purr.backends.qblox.device import (
    QbloxControlHardware,
    QbloxPhysicalBaseband,
    QbloxPhysicalChannel,
)
from qat.purr.backends.qblox.ir import Sequence
from qat.purr.backends.qblox.live import QbloxLiveHardwareModel
from qat.purr.compiler.devices import ChannelType

# TODO: 32Q support: COMPILER-728
_DUMMY_CONFIG = {
    1: ClusterType.CLUSTER_QCM_RF,
    2: ClusterType.CLUSTER_QCM_RF,
    3: ClusterType.CLUSTER_QCM_RF,
    4: ClusterType.CLUSTER_QCM_RF,
    5: ClusterType.CLUSTER_QCM_RF,
    6: ClusterType.CLUSTER_QCM_RF,
    7: ClusterType.CLUSTER_QCM_RF,
    8: ClusterType.CLUSTER_QCM_RF,
    10: ClusterType.CLUSTER_QCM_RF,
    12: ClusterType.CLUSTER_QCM_RF,
    13: ClusterType.CLUSTER_QRM_RF,
    14: ClusterType.CLUSTER_QRM_RF,
    16: ClusterType.CLUSTER_QRM_RF,
    18: ClusterType.CLUSTER_QRM_RF,
}


class DummyQbloxControlHardware(QbloxControlHardware):
    shot_pattern = regex.compile("jlt( +)R([0-9]+),([0-9]+),@(.*)\n")

    def _setup_dummy_scope_acq_data(self, sequencer: Sequencer, sequence: Sequence):
        shot_match = next(self.shot_pattern.finditer(sequence.program), None)
        avg_count = int(shot_match.group(3)) if shot_match else 1

        dummy_data = np.random.random(
            size=(Constants.MAX_SAMPLE_SIZE_SCOPE_ACQUISITIONS, 2)
        )
        dummy_data = [(iq[0], iq[1]) for iq in dummy_data]
        dummy_scope_acquisition_data = DummyScopeAcquisitionData(
            data=dummy_data, out_of_range=(False, False), avg_cnt=(avg_count, avg_count)
        )
        sequencer.set_dummy_scope_acquisition_data(data=dummy_scope_acquisition_data)

    def _setup_dummy_binned_acq_data(self, sequencer: Sequencer, sequence: Sequence):
        shot_match = next(self.shot_pattern.finditer(sequence.program), None)
        avg_count = int(shot_match.group(3)) if shot_match else 1

        for name, acquisition in sequence.acquisitions.items():
            dummy_binned_acquisition_data = [
                DummyBinnedAcquisitionData(
                    data=(np.random.random(), np.random.random()),
                    thres=np.random.choice(2),
                    avg_cnt=avg_count,
                )
            ] * acquisition["num_bins"]
            sequencer.set_dummy_binned_acquisition_data(
                acq_index_name=name,
                data=dummy_binned_acquisition_data,
            )

    def _delete_acquisitions(self, sequencer):
        sequencer.delete_dummy_scope_acquisition_data()
        sequencer.delete_dummy_binned_acquisition_data()

    def set_data(self, packages: list[QbloxPackage]):
        super().set_data(packages)

        # Stage Scope and Acquisition data
        for pulse_channel_id, sequencer in self._id2seq.items():
            if ChannelType.macq.name in pulse_channel_id:
                package = next(
                    (pkg for pkg in packages if pkg.pulse_channel_id == pulse_channel_id)
                )
                self._setup_dummy_scope_acq_data(sequencer, package.sequence)
                self._setup_dummy_binned_acq_data(sequencer, package.sequence)


def get_default_dummy_control_hardware() -> DummyQbloxControlHardware:
    name = "default_dummy_control"
    return DummyQbloxControlHardware(dev_id=name, name=name, dummy_cfg=_DUMMY_CONFIG)


def _letter_from_int(val: int) -> str:
    letter = chr(ord("A") + (val % 25))
    if val >= 25:
        letter = _letter_from_int(val // 25 - 1) + letter

    return letter


def _get_module_slot_key_iterators(instrument: DummyQbloxControlHardware):
    cluster = instrument.driver
    qcmrfs = iter(
        cluster.get_connected_modules(
            filter_fn=lambda mod: mod.is_qcm_type and mod.is_rf_type
        ).keys()
    )
    qrmrfs = iter(
        cluster.get_connected_modules(
            filter_fn=lambda mod: mod.is_qrm_type and mod.is_rf_type
        ).keys()
    )

    return qcmrfs, qrmrfs


def create_qcmrf_qrmrf_config():
    qcmrf_config = QbloxConfig(
        module=ModuleConfig(
            lo=LoConfig(out0_en=True, out1_en=True),
            attenuation=AttConfig(out0=0, out1=0),
            offset=OffsetConfig(out0_path0=0, out0_path1=0, out1_path0=0, out1_path1=0),
        ),
        sequencers={
            i: SequencerConfig(
                connection=ConnectionConfig(bulk_value=[f"out{i % 2}"]),
            )
            for i in range(6)
        },
    )

    qrmrf_config = QbloxConfig(
        module=ModuleConfig(lo=LoConfig(out0_in0_en=True)),
        sequencers={
            i: SequencerConfig(
                connection=ConnectionConfig(bulk_value=["out0", "in0"]),
            )
            for i in range(6)
        },
    )

    return qcmrf_config, qrmrf_config


def apply_setup_to_hardware(
    model: QbloxLiveHardwareModel,
    qubit_count: int = 2,
    connectivity: Connectivity | list[tuple[int, int]] | None = None,
    add_direction_couplings: bool = False,
    instrument: DummyQbloxControlHardware = None,
):
    instrument = instrument or get_default_dummy_control_hardware()
    disconnect_after = False
    if not instrument.is_connected:
        instrument.connect()
        disconnect_after = True

    qcmrfs, qrmrfs = _get_module_slot_key_iterators(instrument)
    qcmrf_config, qrmrf_config = create_qcmrf_qrmrf_config()

    for index in range(qubit_count):
        if index % 2 == 0:
            qcmrf_slot = next(qcmrfs)
        if index % 6 == 0:
            qrmrf_slot = next(qrmrfs)

        letter = _letter_from_int(index)

        bb_q = QbloxPhysicalBaseband(
            f"{letter}-LO-0-QCM-RF-{qcmrf_slot}",
            4.024e9,
            250e6,
            instrument=instrument,
            slot_idx=qcmrf_slot,
            config=qcmrf_config,
        )

        bb_r = QbloxPhysicalBaseband(
            f"{letter}-LO-0-QRM-RF-{qrmrf_slot}",
            9.7772e9,
            250e6,
            instrument=instrument,
            slot_idx=qrmrf_slot,
            config=qrmrf_config,
        )

        ch_q = QbloxPhysicalChannel(f"{letter}-CH-QCM-RF-{qcmrf_slot}", 1e-9, bb_q, 1)
        ch_r = QbloxPhysicalChannel(
            f"{letter}-CH-QRM-RF-{qrmrf_slot}", 1e-9, bb_r, 1, acquire_allowed=True
        )

        resonator = ch_r.build_resonator(f"R{index}", frequency=10.2033e9)
        qubit = ch_q.build_qubit(
            index,
            resonator,
            drive_freq=3.872e9,
            second_state_freq=4.085e9,
            channel_scale=1,
            measure_amp=0.5,
            fixed_drive_if=True,
        )
        # TODO - Add support for mapping cross resonance channels to the right QBlox sequencers
        # add_cross_resonance(qubit, resonator)
        # COMPILER-685

        model.add_physical_baseband(bb_q, bb_r)
        model.add_physical_channel(ch_q, ch_r)
        model.add_quantum_device(resonator, qubit)
        model.control_hardware = instrument

    if disconnect_after:
        instrument.disconnect()
    return model


def get_default_dummy_hardware(
    qubit_count: int = 2,
    connectivity: Connectivity | list[tuple[int, int]] | None = None,
    add_direction_couplings: bool = False,
    name: str | None = None,
    dummy_cfg: dict | None = None,
):
    name = name or f"dummy_model_{qubit_count}_{uuid.uuid4()}".replace("-", "_")
    instrument = DummyQbloxControlHardware(
        dev_id=name, name=name, dummy_cfg=dummy_cfg or _DUMMY_CONFIG
    )

    model = QbloxLiveHardwareModel()
    apply_setup_to_hardware(
        model,
        qubit_count=qubit_count,
        connectivity=connectivity,
        add_direction_couplings=add_direction_couplings,
        instrument=instrument,
    )

    return model



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd


from compiler_config.config import Languages
from qiskit import qasm2
from qiskit.circuit import (
    Barrier,
    ClassicalRegister,
    Delay,
    Gate,
    IfElseOp,
    Measure,
    QuantumRegister,
    Reset,
)
from qiskit.circuit.library import CXGate, UGate
from qiskit.converters import circuit_to_dag
from qiskit.dagcircuit import DAGCircuit, DAGInNode, DAGNode, DAGOpNode, DAGOutNode
from qiskit.qasm2 import CustomInstruction

from qat.frontend.parsers.qasm.base import AbstractParser, QasmContext
from qat.ir.instruction_builder import QuantumInstructionBuilder
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class Qasm2Parser(AbstractParser):
    ecr_gate = Gate("ecr", 2, [])

    def __init__(self, order_result_vars=False, raw_results=False):
        super().__init__()
        self.order_result_vars = order_result_vars
        self.raw_results = raw_results
        self._cached_parses: dict[int, DAGCircuit] = dict()

    def __repr__(self):
        return self.__class__.__name__

    def parser_language(self) -> Languages:
        return Languages.Qasm2

    def _get_intrinsics(self, qasm="") -> list[CustomInstruction]:
        instrinsics = list(qasm2.LEGACY_CUSTOM_INSTRUCTIONS)
        if "gate ecr" not in qasm:
            instrinsics.append(
                qasm2.CustomInstruction("ecr", 0, 2, lambda: self.ecr_gate, builtin=True)
            )
        return instrinsics

    def _fetch_or_parse(self, qasm: str) -> DAGCircuit:
        # If we have seen this file before
        qasm_id = hash(qasm)
        if (cached_value := self._cached_parses.get(qasm_id, None)) is not None:
            return cached_value

        circ = qasm2.loads(qasm, custom_instructions=self._get_intrinsics(qasm))
        program = circuit_to_dag(circ)
        from importlib.metadata import version

        if version("qiskit") < "2.0.0":
            # TODO: Drop conversion when we enforce Qiskit 2.x - COMPILER-658
            from qiskit.transpiler.passes.utils.convert_conditions_to_if_ops import (
                ConvertConditionsToIfOps,
            )

            program = ConvertConditionsToIfOps().run(program)

        self._cached_parses[qasm_id] = program
        return program

    def parse(
        self, builder: QuantumInstructionBuilder, qasm: str
    ) -> QuantumInstructionBuilder:
        # Parse or pick up the cached version, then remove it as we are about to
        # interpret it.
        program = self._fetch_or_parse(qasm)
        if (qasm_id := hash(qasm)) in self._cached_parses:
            del self._cached_parses[qasm_id]
        return self.process_program(builder, program)

    def validate(self, circ: DAGCircuit):
        pass

    def modify(self, circ: DAGCircuit):
        """
        Allows children to transform the program before validation/transforming into our
        AST occurs.
        """
        pass

    def _walk_program(
        self, builder: QuantumInstructionBuilder, circ: DAGCircuit, context: QasmContext
    ):
        self.modify(circ)
        self.validate(circ)
        self._current_dag = circ

        for i, node in enumerate(circ.nodes()):
            self.walk_node(node, context, builder)

    def process_program(
        self, builder: QuantumInstructionBuilder, circ: DAGCircuit
    ) -> QuantumInstructionBuilder:
        context = QasmContext()
        self._walk_program(builder, circ, context)

        # 'raw' results in this case simply means return the base measurement array in
        # the order of execution.
        register_keys = context.registers.classic.keys()
        if self.order_result_vars:
            register_keys = sorted(register_keys)

        for key in register_keys:
            builder.assign(key, [val.value for val in context.registers.classic[key].bits])

        builder.returns([key for key in register_keys])
        return builder

    def process_gate(
        self,
        node: DAGOpNode,
        context: QasmContext,
        builder: QuantumInstructionBuilder,
        **kwargs,
    ):
        """Process a gate call."""
        if not isinstance(node.op, Gate):
            raise ValueError(
                f"Node {str(node)} is not actually a gate but being treated like one."
            )

        gate_def = node.op.definition
        if gate_def is None or len(gate_nodes := circuit_to_dag(gate_def).gate_nodes()) < 1:
            self.process_intrinsic(node, context, builder)
            return
        for gate_node in gate_nodes:
            qs = tuple([node.qargs[gate_def.qubits.index(q)] for q in gate_node.qargs])
            gate_node.qargs = qs
            self.walk_node(gate_node, context, builder)

    def process_intrinsic(
        self, node: DAGOpNode, context: QasmContext, builder: QuantumInstructionBuilder
    ):
        # Only process if it's an ECR gate and, more pertinently, our ECR gate.
        if node.name == "ecr":
            # TODO: Define precisely what rotation this should be, including argument.
            qubits = self._get_qubits(node, context)
            self._add_ecr(qubits, builder)
        else:
            raise ValueError(f"Gate {node.name} isn't intrinsic and has no body.")

    def process_barrier(
        self,
        node: DAGOpNode,
        context: QasmContext,
        builder: QuantumInstructionBuilder,
        **kwargs,
    ):
        pass

    def process_delay(
        self,
        node: DAGOpNode,
        context: QasmContext,
        builder: QuantumInstructionBuilder,
        **kwargs,
    ):
        qubits = self._get_qubits(node, context)
        delay = node.op.duration
        self._add_delay(delay, qubits, builder)

    def process_gate_definition(self, node: DAGOpNode, context: QasmContext, _, **kwargs):
        context.gates[node.name] = node

    def process_creg(
        self,
        node: ClassicalRegister,
        context: QasmContext,
        builder: QuantumInstructionBuilder,
        **kwargs,
    ):
        self._add_creg(node.name, node.size, context)

    def process_qreg(
        self,
        node: QuantumRegister,
        context: QasmContext,
        builder: QuantumInstructionBuilder,
        **kwargs,
    ):
        self._add_qreg(node.name, node.size, context, builder)

    def process_measure(
        self,
        node: DAGOpNode,
        context: QasmContext,
        builder: QuantumInstructionBuilder,
        **kwargs,
    ):
        self._add_measure(
            self._get_qubits(node, context),
            self._get_clbits(node, context),
            builder,
        )

    def process_unitary(
        self,
        node: DAGOpNode,
        context: QasmContext,
        builder: QuantumInstructionBuilder,
        **kwargs,
    ):
        """Unitary in QASM terms is just ``U(...)``."""
        theta, phi, _lambda = node.op.params
        self._add_unitary(theta, phi, _lambda, self._get_qubits(node, context), builder)

    def process_cnot(
        self,
        node: DAGOpNode,
        context: QasmContext,
        builder: QuantumInstructionBuilder,
        **kwargs,
    ):
        self._add_cnot(
            *self._get_qubits(node, context),
            builder,
        )

    def process_reset(
        self,
        node: DAGOpNode,
        context: QasmContext,
        builder: QuantumInstructionBuilder,
        **kwargs,
    ):
        self._add_reset(self._get_qubits(node, context), builder)

    def walk_node(
        self,
        node: DAGNode,
        context: QasmContext,
        builder: QuantumInstructionBuilder,
        **kwargs,
    ):
        """
        Process each individual QASM node, builds context or forwards processing to
        relevant ``process_x`` method associated with each node type.
        """
        if isinstance(node, (DAGInNode, DAGOutNode)):
            for register, _ in self._current_dag.find_bit(node.wire).registers:
                if isinstance(register, QuantumRegister):
                    self.process_qreg(register, context, builder)
                elif isinstance(register, ClassicalRegister):
                    self.process_creg(register, context, builder)
        elif isinstance(node, DAGOpNode):
            match node.op:
                case Delay():
                    self.process_delay(node, context, builder, **kwargs)
                case Barrier():
                    self.process_barrier(node, context, builder, **kwargs)
                case Measure():
                    self.process_measure(node, context, builder, **kwargs)
                case Reset():
                    self.process_reset(node, context, builder, **kwargs)
                case UGate():
                    self.process_unitary(node, context, builder, **kwargs)
                case CXGate():
                    self.process_cnot(node, context, builder, **kwargs)
                case IfElseOp():
                    raise NotImplementedError("IfElseOp is not currently supported.")
                case Gate():
                    self.process_gate(node, context, builder, **kwargs)
                case _:
                    raise NotImplementedError(
                        "DAGNode action not implemented for '{type(node)}'."
                    )

    def _get_parameters(self, node: DAGOpNode, context: QasmContext) -> list:
        """Get the params of a gate. These are the non-qubit values of a gate."""
        if isinstance(node.op, Gate):
            return node.op.params
        if isinstance(node.op, Measure):
            bits = self._get_clbits(node, context)
            if not isinstance(bits, list):
                bits = [bits]

            return bits
        return []

    def _get_qubits(self, node: DAGOpNode, context: QasmContext) -> list:
        """
        Resolve what qubits or qubit registers this node relates too.

        If a value was originally a qubit register, it will be appended as a list,
        otherwise if it was a single qubit target, just a normal object.

        :Example:

        .. code-block:: python

            [[q0, q1, q2], q2]

        The above means the first target was a register referencing q1-3, whereas the
        second was a single qubit referenced directly. If we say our register is just
        called qb, the above would equate to an argument list of ``[qb, qb[2]]``.
        """
        qubits = []
        for qubit in node.qargs:
            reg, ind = self._current_dag.find_bit(qubit).registers[0]
            register = context.registers.quantum.get(reg.name, None)
            if register is not None:
                qubits.append(register.qubits[ind])
        return qubits

    def _get_clbits(self, source: DAGOpNode | tuple, context: QasmContext) -> list:
        if isinstance(source, DAGOpNode):
            return self._get_clbits(source.cargs, context)
        else:
            clbits = []
            for clbit in source:
                reg, ind = self._current_dag.find_bit(clbit).registers[0]
                register = context.registers.classic.get(reg.name, None)
                if register is not None:
                    clbits.append(register.bits[ind])
            return clbits


class RestrictedQasm2Parser(Qasm2Parser):
    """Parser which only allows certain gates to be passed."""

    def __init__(self, allowed_gates=None, disable_if=False, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.allowed_gates = allowed_gates
        self.disable_if = disable_if

    def validate(self, circ: DAGCircuit):
        if self.allowed_gates is not None:
            intrinsic_gates = {node.name: node for node in qasm2.LEGACY_CUSTOM_INSTRUCTIONS}

            # Look at both the main script and custom gate intrinsic usage.
            gate_nodes = {
                val.name for val in circ.gate_nodes() if val.name in intrinsic_gates
            }

            invalid_gates = gate_nodes.difference(self.allowed_gates)
            if any(invalid_gates):
                raise ValueError(
                    f"Gates [{', '.join(invalid_gates)}] are currently unable to be used."
                )

        if self.disable_if and any(
            [node.op for node in circ.op_nodes() if isinstance(node.op, IfElseOp)]
        ):
            raise ValueError("If's are currently unable to be used.")


class CloudQasmParser(RestrictedQasm2Parser):
    """
    QASM parser used in our QCaaS system.
    """

    def __init__(self):
        super().__init__(allowed_gates=None, disable_if=True, order_result_vars=True)



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import json
import os
from dataclasses import asdict
from datetime import datetime

from qat.backend.qblox.acquisition import Acquisition
from qat.backend.qblox.execution import QbloxProgram
from qat.backend.qblox.visualisation import plot_playback, plot_program
from qat.engines import NativeEngine
from qat.instrument.base import InstrumentConcept
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class QbloxEngine(NativeEngine[QbloxProgram]):
    def __init__(self, instrument: InstrumentConcept):
        self.instrument: InstrumentConcept = instrument

        self.plot_program = False
        self.dump_program = False
        self.plot_playback = False

    def execute(self, program: QbloxProgram, **kwargs) -> dict[str, list[Acquisition]]:
        if self.plot_program:
            plot_program(program)

        if self.dump_program:
            for pulse_channel_id, pkg in program.packages.items():
                filename = f"schedules/target_{pulse_channel_id}_{datetime.now().strftime('%m-%d-%Y_%H%M%S')}.json"
                os.makedirs(os.path.dirname(filename), exist_ok=True)
                with open(filename, "w") as f:
                    f.write(json.dumps(asdict(pkg.sequence)))

        self.instrument.setup(program)
        playback: dict[str, list[Acquisition]] = self.instrument.playback()

        if self.plot_playback:
            plot_playback(playback)

        return playback



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from qat.core.pass_base import ValidationPass
from qat.ir.instruction_builder import InstructionBuilder
from qat.ir.measure import Acquire


class NoAcquireWeightsValidation(ValidationPass):
    """Some target machines do not support :class:`Acquire` instructions that contain weights.
    This pass can be used to validate that this is the case."""

    def run(self, ir: InstructionBuilder, *args, **kwargs):
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        """

        has_filters = [inst.filter for inst in ir.instructions if isinstance(inst, Acquire)]
        if any(has_filters):
            raise NotImplementedError(
                "Acquire filters are not implemented for this target machine."
            )
        return ir


PydNoAcquireWeightsValidation = NoAcquireWeightsValidation



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from compiler_config.config import CompilerConfig, MetricsType

from qat.backend.base import BaseBackend
from qat.core.metrics_base import MetricsManager
from qat.core.result_base import ResultManager
from qat.engines import NativeEngine
from qat.executables import Executable
from qat.frontend import BaseFrontend
from qat.middleend.base import BaseMiddleend
from qat.model.target_data import TargetData
from qat.model.validators import MismatchingHardwareModelException
from qat.pipelines.base import BasePipeline
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.runtime.base import BaseRuntime


class CompilePipeline(BasePipeline):
    """Implements a pipeline that compiles quantum programs.

    In addition to the hardware model and target data, a compilation pipeline consists of
    the following components:

    #. Frontend: Compiles a high-level language-specific, but target-agnostic,
       input (e.g., QASM, QIR, ...) to QAT's intermediate representation (IR), QatIR.
    #. Middleend: Takes the QatIR and performs a sequences of  passes that validate and
       optimise the IR, and prepare it for codegen.
    #. Backend: Handles code generation to allow the program to be executed on the target.

    In the future, this might be relaxed so compilation pipelines can be defined more
    abstractly, only requiring the :meth:`compile` to be implemented.
    """

    def __init__(
        self,
        name: str,
        model: QuantumHardwareModel,
        frontend: BaseFrontend,
        middleend: BaseMiddleend,
        backend: BaseBackend,
        target_data: TargetData | None = None,
        disable_model_validation: bool = False,
    ):
        if not disable_model_validation:
            self._validate_consistent_model(model, frontend, middleend, backend)
        self._name = name
        self._model = model
        self._target_data = target_data if target_data is not None else TargetData.default()
        self._frontend = frontend
        self._middleend = middleend
        self._backend = backend
        self.disable_model_validation = disable_model_validation

    def compile(
        self,
        program,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ) -> tuple[Executable, MetricsManager]:
        """Compiles a source program into an executable using the pipeline's components.

        :param program: The source program to compile.
        :param compiler_config: Configuration options for the compiler, such as optimization
            and results formatting.
        :return: An executable of the compiled program for the target device and the metrics
            manager containing metrics collected during compilation.
        """

        # TODO: Improve metrics and config handling
        compiler_config = compiler_config or CompilerConfig(
            metrics=MetricsType.Experimental
        )
        metrics_manager = MetricsManager(compiler_config.metrics)
        compilation_results = ResultManager()

        ir = self.frontend.emit(
            program,
            compilation_results,
            metrics_manager,
            compiler_config=compiler_config,
            **kwargs,
        )

        ir = self.middleend.emit(
            ir,
            compilation_results,
            metrics_manager,
            compiler_config=compiler_config,
            **kwargs,
        )

        package = self.backend.emit(
            ir,
            compilation_results,
            metrics_manager,
            compiler_config=compiler_config,
            **kwargs,
        )

        return package, metrics_manager

    @property
    def frontend(self) -> BaseFrontend:
        return self._frontend

    @property
    def middleend(self) -> BaseMiddleend:
        return self._middleend

    @property
    def backend(self) -> BaseBackend:
        return self._backend


class ExecutePipeline(BasePipeline):
    """Implements a pipeline that can execute quantum programs.

    In addition to the hardware model and target data, an execution pipeline consists of
    the following components:

    #. Runtime: Manages the execution of the program, including the engine and the post-
       processing of the results.
    #. Engine: Communicates the compiled program with the target devices, and returns the
       results.

    In the future, this might be relaxed so execution pipelines can be defined more
    abstractly, only requiring the :meth:`execute` to be implemented.
    """

    def __init__(
        self,
        name: str,
        model: QuantumHardwareModel,
        runtime: BaseRuntime,
        target_data: TargetData | None = None,
        disable_model_validation: bool = False,
    ):
        if not disable_model_validation:
            self._validate_consistent_model(model, runtime, runtime.engine)
        super().__init__(name, model, target_data)
        self._runtime = runtime
        self.disable_model_validation = disable_model_validation

    def execute(
        self,
        executable: Executable,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ) -> tuple[dict, MetricsManager]:
        """Uses the runtime and engine in the pipeline to execute a compiled program and
        process the results.

        Checks that the hardware model in the pipeline matches the hardware model used
        during compilation.

        :param executable: The compiled program to execute.
        :param compiler_config: Configuration options for the compiler, such as optimization
            and results formatting.
        :return: A dictionary of results from the execution and the metrics manager
            containing metrics collected during execution.
        """

        compiler_config = CompilerConfig() if compiler_config is None else compiler_config

        if (
            self.model is not None
            and self.model.calibration_id != executable.calibration_id
        ):
            raise MismatchingHardwareModelException(
                f"Hardware id '{executable.calibration_id}' in the executable "
                f"does not match the hardware model id '{self.model.calibration_id}' used "
                "for execution in this pipeline."
            )

        pp_results = ResultManager()
        metrics_manager = MetricsManager(compiler_config.metrics)
        results = self.runtime.execute(
            executable,
            res_mgr=pp_results,
            met_mgr=metrics_manager,
            compiler_config=compiler_config,
            **kwargs,
        )
        return results, metrics_manager

    @property
    def runtime(self) -> BaseRuntime:
        return self._runtime

    @property
    def engine(self) -> NativeEngine:
        """Returns the engine within the runtime."""
        return self._runtime.engine


class Pipeline(CompilePipeline, ExecutePipeline):
    """A composite pipeline that can both compile and execute quantum programs.

    This pipeline combines the functionality of both compilation and execution pipelines,
    allowing for a complete workflow from source program to execution. See
    :class:`CompilePipeline` and :class:`ExecutePipeline` for more details on the
    components involved in compilation and execution.
    """

    def __init__(
        self,
        name: str,
        model: QuantumHardwareModel,
        frontend: BaseFrontend,
        middleend: BaseMiddleend,
        backend: BaseBackend,
        runtime: BaseRuntime,
        target_data: TargetData | None = None,
        disable_model_validation: bool = False,
    ):
        if not disable_model_validation:
            self._validate_consistent_model(
                model, frontend, middleend, backend, runtime, runtime.engine
            )
        self._name = name
        self._model = model
        self._target_data = target_data if target_data is not None else TargetData.default()
        self._frontend = frontend
        self._middleend = middleend
        self._backend = backend
        self._runtime = runtime
        self.disable_model_validation = disable_model_validation



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from collections import defaultdict
from itertools import chain

from qat.core.pass_base import LoweringPass
from qat.ir.lowered import PartitionedIR
from qat.middleend.passes.purr.transform import LoopCount
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.instructions import (
    Acquire,
    Assign,
    GreaterThan,
    InlineResultsProcessing,
    Instruction,
    Jump,
    Label,
    PostProcessing,
    QuantumInstruction,
    Repeat,
    ResultsProcessing,
    Return,
    Synchronize,
    Variable,
)


class PartitionByPulseChannel(LoweringPass):
    """Separates a list of instructions into their respective pulse channels.

    For targets that do not have native support for
    :class:`Synchronize <qat.purr.compiler.instructions.Synchronize>`, we can separate the
    instruction list into many lists of instructions, with each list only containing
    instructions that act on a single channel.

    This pass achieves the same as the :class:`TriagePass <qat.backend.passes.legacy.analysis.TriagePass>`,
    but instead of saving the results in the results manager, returns the partitioned
    instructions. The only reason I've changed the name is to not confuse it with the
    TriagePass :)
    """

    def run(self, ir: InstructionBuilder, *args, **kwargs) -> PartitionedIR:
        """
        :param ir: The list of instructions stored in an :class:`InstructionBuilder`.
        :param res_mgr: The result manager to save the analysis results.
        """
        shared_instructions = list()
        variables = defaultdict(list)
        partitioned_ir = PartitionedIR()

        if hasattr(ir, "shots") and hasattr(ir, "compiled_shots"):
            partitioned_ir.shots = ir.shots
            partitioned_ir.compiled_shots = ir.compiled_shots

        partitioned_ir.repetition_period = getattr(ir, "repetition_period", None)

        partitioned_ir.passive_reset_time = getattr(ir, "passive_reset_time", None)

        def add_to_all(inst: Instruction) -> None:
            shared_instructions.append(inst)
            for target_list in partitioned_ir.target_map.values():
                target_list.append(inst)
            partitioned_ir.target_map.default_factory = shared_instructions.copy

        repeat_seen = False
        for inst in ir.instructions:
            handled = False
            if isinstance(inst, QuantumInstruction) and not isinstance(
                inst, PostProcessing
            ):
                if isinstance(inst, Synchronize):
                    raise ValueError(
                        f"The `Synchronize` instruction {inst} is not supported by the "
                        "PartitionByPulseChannelPass. Please lower it to `Delay` instructions "
                        "using the `LowerSyncsToDelays` pass"
                    )
                handled = True
                for qt in inst.quantum_targets:
                    partitioned_ir.target_map[qt].append(inst)

            if isinstance(inst, Return):
                partitioned_ir.returns.append(inst)

            elif isinstance(inst, Variable):
                variables[inst.name].append(inst)
                add_to_all(inst)

            elif isinstance(inst, Assign):
                var_refs = variables[inst.name]
                if len(var_refs) == 0:
                    variables[inst.name].append(Variable(inst.name))
                if var_refs[-1].var_type == LoopCount:
                    add_to_all(inst)
                if isinstance(inst.value, list | str):
                    # Only list assigns that are readout/result assignements
                    # TODO: Supper different assigns for result readout vs loop
                    # variables etc.
                    partitioned_ir.assigns.append(inst)

            elif isinstance(inst, Acquire):
                for t in inst.quantum_targets:
                    partitioned_ir.acquire_map[t].append(inst)

            elif isinstance(inst, PostProcessing):
                partitioned_ir.pp_map[inst.output_variable].append(inst)

            elif isinstance(inst, ResultsProcessing):
                partitioned_ir.rp_map[inst.variable] = inst

            elif isinstance(inst, Repeat):
                if repeat_seen:
                    raise ValueError("Multiple Repeat instructions found.")
                if partitioned_ir.shots is None:
                    partitioned_ir.shots = inst.repeat_count
                    partitioned_ir.compiled_shots = inst.repeat_count
                if partitioned_ir.repetition_period is None:
                    partitioned_ir.repetition_period = inst.repetition_period
                if partitioned_ir.passive_reset_time is None:
                    partitioned_ir.passive_reset_time = inst.passive_reset_time
                repeat_seen = True

            elif isinstance(inst, Jump):
                if isinstance(inst.condition, GreaterThan):
                    if repeat_seen:
                        raise ValueError("Multiple Repeat or Jump instructions found.")

                    if partitioned_ir.shots is None:
                        limit = inst.condition.left
                        partitioned_ir.shots = limit
                        partitioned_ir.compiled_shots = limit

                add_to_all(inst)

            elif isinstance(inst, Label):
                add_to_all(inst)
            elif not handled:
                raise TypeError(f"Unexpected Instruction type: {type(inst)}.")

        # Assume that raw acquisitions are experiment results.
        # TODO: separate as ResultsProcessingSanitisation pass. (COMPILER-412)
        acquires = list(chain(*partitioned_ir.acquire_map.values()))
        missing_results = {
            acq.output_variable
            for acq in acquires
            if acq.output_variable not in partitioned_ir.rp_map
        }
        for missing_var in missing_results:
            partitioned_ir.rp_map[missing_var] = ResultsProcessing(
                missing_var, InlineResultsProcessing.Experiment
            )

        return partitioned_ir



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from numbers import Number
from typing import Iterable

import numpy as np

# TODO: these functions might benefit from a little attention, and also add tests.
# they are just extracted from various `purr` packages


def binary_average(results_list):
    """
    Extracted from `purr.compiler.execution`.

    Averages all repeat results and returns a definitive 1/0 for each qubit measurement.
    """
    # If we have many sweeps/repeats loop through all of them and sum.
    if all([isinstance(val, list) for val in results_list]):
        binary_results = [binary_average(nested) for nested in results_list]
    else:
        binary_results = binary(results_list)

    return 1 if sum(binary_results) >= (len(binary_results) / 2) else 0


def binary(results_list):
    """
    Extracted from `purr.compiler.execution`.

    Changes all measurements to binary format.
    """
    if not isinstance(results_list, Iterable):
        return [results_list]

    results = []
    for item in results_list:
        if isinstance(item, Iterable):
            results.append(binary(item))
        elif isinstance(item, complex):  # If we're a flat register, just append.
            results.append(complex_to_binary(item))
        elif isinstance(item, float):
            results.append(0 if item > 0 else 1)
        else:
            results.append(item)
    return results


def complex_to_binary(number: complex):
    """
    Extracted from `purr.compiler.execution`.

    Base calculation for changing a complex measurement to binary form.
    """
    return 0 if number.real > 0 else 1


def numpy_array_to_list(array):
    """
    Extracted from `purr.compiler.execution`.

    Transform numpy arrays to a normal list.
    """
    if isinstance(array, np.ndarray):
        numpy_list = array.tolist()
        if len(numpy_list) == 1:
            return numpy_list[0]
        return numpy_list
    elif isinstance(array, list):
        list_list = [numpy_array_to_list(val) for val in array]
        if len(list_list) == 1:
            return list_list[0]
        return list_list
    else:
        return array


def binary_count(results_list, repeats):
    """
    Extracted from `qat.purr.compiler.runtime`.

    Returns a dictionary of binary number: count. So for a two qubit register it'll return
        the various counts for ``00``, ``01``, ``10`` and ``11``.
    """

    def flatten(res):
        """
        Combine binary result from the QPU into composite key result.
        Aka '0110' or '0001'
        """
        if isinstance(res, Iterable):
            return "".join([flatten(val) for val in res])
        else:
            return str(res)

    def get_tuple(res, index):
        return [val[index] if isinstance(val, (list, np.ndarray)) else val for val in res]

    binary_results = binary(results_list)

    # If our results are a single qubit then pretend to be a register of one.
    if (
        isinstance(next(iter(binary_results), None), Number)
        and len(binary_results) == repeats
    ):
        binary_results = [binary_results]

    result_count = dict()
    for qubit_result in [list(get_tuple(binary_results, i)) for i in range(repeats)]:
        key = flatten(qubit_result)
        value = result_count.get(key, 0)
        result_count[key] = value + 1

    return result_count



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from warnings import warn

from qat.pipelines.base import AbstractPipeline
from qat.pipelines.pipeline import ExecutePipeline


def ExecuteSweepPipeline(base_pipeline: AbstractPipeline):
    """Runtimes now support execution of batched programs in an executable, so the old
    ExecuteSweepPipeline is no longer necessary. This function is kept for backwards
    compatibility, and simply returns the base pipeline after checking its type."""

    warn(
        "The behaviour of ExecuteSweepPipeline is now supported the SimpleRuntime, and an "
        "expclicit ExecuteSweepPipeline is no longer required. This is considered "
        "and will be removed in a future release. ",
        DeprecationWarning,
        stacklevel=2,
    )

    if not base_pipeline.is_subtype_of(ExecutePipeline):
        raise TypeError("The base pipeline must be an ExecutePipeline.")
    return base_pipeline



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from functools import singledispatchmethod

from qat.ir.instruction_builder import (
    QuantumInstructionBuilder as PydQuantumInstructionBuilder,
)
from qat.model.hardware_model import PhysicalHardwareModel
from qat.purr.compiler.builders import (
    QuantumInstructionBuilder as PurrQuantumInstructionBuilder,
)
from qat.purr.compiler.hardware_models import QuantumHardwareModel


class BuilderFactory:
    @singledispatchmethod
    @staticmethod
    def create_builder(model):
        raise TypeError(
            f"Cannot find a builder for hardware model with type {type(model)}."
        )

    @create_builder.register
    @staticmethod
    def _(model: PhysicalHardwareModel) -> PydQuantumInstructionBuilder:
        return PydQuantumInstructionBuilder(model)

    @create_builder.register
    @staticmethod
    def _(model: QuantumHardwareModel) -> PurrQuantumInstructionBuilder:
        return model.create_builder()



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import re
from dataclasses import dataclass

from qat.core.pass_base import AnalysisPass
from qat.core.result_base import ResultInfoMixin, ResultManager
from qat.executables import Executable
from qat.ir.instruction_builder import InstructionBuilder
from qat.ir.measure import Acquire
from qat.model.hardware_model import PhysicalHardwareModel


@dataclass
class IndexMappingResult(ResultInfoMixin):
    mapping: dict[str, int]


class IndexMappingAnalysis(AnalysisPass):
    """Determines a mapping from classical bit registers to qubits.

    Searches through the acquires of the package and determines their associated qubit.
    Also looks for classical registers of the form :code:`<clreg_name>[<clreg_index>]`.

    Supports both :class:`Executable` packages and :class:`InstructionBuilder`.
    """

    def __init__(self, model: PhysicalHardwareModel):
        """:param model: The hardware model is needed for the qubit mapping."""

        self.model = model
        pulse_to_phys_channel_map: dict[str, str] = {}
        for qubit in self.model.qubits.values():
            for pulse_ch in qubit.all_pulse_channels:
                pulse_to_phys_channel_map[pulse_ch.uuid] = qubit.physical_channel.uuid

            for pulse_ch in qubit.resonator.all_pulse_channels:
                pulse_to_phys_channel_map[pulse_ch.uuid] = (
                    qubit.resonator.physical_channel.uuid
                )
        self.pulse_to_phys_channel_map = pulse_to_phys_channel_map

    def run(
        self,
        acquisitions: dict[str, any],
        res_mgr: ResultManager,
        *args,
        package: InstructionBuilder | Executable,
        **kwargs,
    ):
        """
        :param acquisitions: The dictionary of results acquired from the target machine.
        :param res_mgr: The results manager to save the mapping.
        :param package: The executable program containing the results-processing
            information should be passed as a keyword argument.
        """
        # Determine a mapping from output variable - > qubit index.
        if isinstance(package, Executable):
            var_to_phys_channel_map = self.var_to_physical_channel_executable(package)
        else:
            var_to_phys_channel_map = self.var_to_physical_channel_qat_ir(package)
        var_to_qubit_map = self.var_to_qubit_map(var_to_phys_channel_map)

        # Search for classical registers defined in e.g. QASM.
        pattern = re.compile(r"(.*)\[([0-9]+)\]")
        for var in list(var_to_qubit_map.keys()):
            result = pattern.match(var)
            if result:
                var_to_qubit_map[result.group(2)] = var_to_qubit_map.pop(var)

        res_mgr.add(IndexMappingResult(mapping=var_to_qubit_map))
        return acquisitions

    @staticmethod
    def var_to_physical_channel_executable(package: Executable) -> dict[str, str]:
        return {var: acquire.physical_channel for var, acquire in package.acquires.items()}

    def var_to_physical_channel_qat_ir(self, package: InstructionBuilder) -> dict[str, str]:
        mapping: dict[str, str] = {}
        for instruction in package.instructions:
            if isinstance(instruction, Acquire):
                mapping[instruction.output_variable] = self.pulse_to_phys_channel_map[
                    instruction.target
                ]
        return mapping

    def var_to_qubit_map(self, mapping: dict[str, str]):
        """
        Maps the variables in the mapping to qubit indices.
        """

        var_to_qubit_map: dict[str, int] = {}
        for var, phys_ch_id in mapping.items():
            acquired_qubits = [
                idx
                for idx, qubit in self.model.qubits.items()
                if qubit.resonator.physical_channel.uuid == phys_ch_id
            ]

            if len(acquired_qubits) == 0:
                raise ValueError(
                    f"Could not find any qubits with acquire channel {phys_ch_id}."
                )
            var_to_qubit_map[var] = acquired_qubits[0]

        return var_to_qubit_map



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from typing import Callable

from qat.core.config.descriptions import (
    CompilePipelineDescription,
    ExecutePipelineDescription,
    PipelineClassDescription,
)
from qat.core.config.validators import requires_model, requires_target_data
from qat.engines import NativeEngine
from qat.model.target_data import AbstractTargetData
from qat.pipelines.pipeline import CompilePipeline, ExecutePipeline, Pipeline
from qat.pipelines.updateable import Model, UpdateablePipeline


def _inject_model_and_target_data(
    func: Callable | None, model: Model, target_data: AbstractTargetData, **kwargs
):
    """Injects the model and target data into the function if required.

    Hopefully this will eventually be replaced with a dependency injection solution.
    """
    if func is None:
        return None
    if requires_model(func):
        kwargs["model"] = model
    if requires_target_data(func):
        kwargs["target_data"] = target_data
    return func(**kwargs)


class ConfigurableCompilePipeline(UpdateablePipeline):
    """A pipeline that allows each constituent to be defined in a granular fashion, and
    updated. The config is provided via the :class:`CompilePipelineDescription`.

    Allows a config to be provided that configures each component of the pipeline via
    the qatconfig.
    """

    @staticmethod
    def _build_pipeline(
        config: CompilePipelineDescription,
        model: Model,
        target_data: AbstractTargetData,
        engine: NativeEngine | None = None,
    ) -> CompilePipeline:
        """Constructs a pipeline with each component defined in the config in a granular
        fashion."""

        frontend = _inject_model_and_target_data(config.frontend, model, target_data)
        middleend = _inject_model_and_target_data(config.middleend, model, target_data)
        backend = _inject_model_and_target_data(config.backend, model, target_data)

        return CompilePipeline(
            name=config.name,
            model=model,
            frontend=frontend,
            middleend=middleend,
            backend=backend,
            target_data=target_data,
        )


class ConfigurableExecutePipeline(UpdateablePipeline):
    """A pipeline that allows each constituent to be defined in a granular fashion, and
    updated. The config is provided via the :class:`ExecutePipelineDescription`.

    Allows a config to be provided that configures each component of the pipeline via
    the qatconfig.
    """

    @staticmethod
    def _build_pipeline(
        config: ExecutePipelineDescription,
        model: Model,
        target_data: AbstractTargetData,
        engine: NativeEngine,
    ) -> Pipeline:
        """Constructs a pipeline with each component defined in the config in a granular
        fashion."""

        results_pipeline = _inject_model_and_target_data(
            config.results_pipeline, model, target_data
        )
        runtime = _inject_model_and_target_data(
            config.runtime,
            model,
            target_data,
            engine=engine,
            results_pipeline=results_pipeline,
        )
        return ExecutePipeline(
            name=config.name,
            model=model,
            runtime=runtime,
            target_data=target_data,
        )


class ConfigurablePipeline(UpdateablePipeline):
    """A pipeline that allows each constituent to be defined in a granular fashion, and
    updated. The config is provided via the :class:`PipelineClassDescription`.

    Allows a config to be provided that configures each component of the pipeline via
    the qatconfig.
    """

    @staticmethod
    def _build_pipeline(
        config: PipelineClassDescription,
        model: Model,
        target_data: AbstractTargetData,
        engine: NativeEngine,
    ) -> Pipeline:
        """Constructs a pipeline with each component defined in the config in a granular
        fashion."""

        frontend = _inject_model_and_target_data(config.frontend, model, target_data)
        middleend = _inject_model_and_target_data(config.middleend, model, target_data)
        backend = _inject_model_and_target_data(config.backend, model, target_data)
        results_pipeline = _inject_model_and_target_data(
            config.results_pipeline, model, target_data
        )
        runtime = _inject_model_and_target_data(
            config.runtime,
            model,
            target_data,
            engine=engine,
            results_pipeline=results_pipeline,
        )

        return Pipeline(
            name=config.name,
            model=model,
            frontend=frontend,
            middleend=middleend,
            backend=backend,
            runtime=runtime,
            target_data=target_data,
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from pydantic import field_validator

from qat.core.pass_base import PassManager
from qat.model.target_data import TargetData
from qat.pipelines.pipeline import ExecutePipeline
from qat.pipelines.updateable import PipelineConfig, UpdateablePipeline
from qat.purr.backends.live import LiveHardwareModel
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.utils.logger import get_default_logger
from qat.runtime.connection import ConnectionMode
from qat.runtime.legacy import LegacyRuntime
from qat.runtime.passes.purr.analysis import IndexMappingAnalysis
from qat.runtime.passes.transform import ErrorMitigation, ResultTransform

log = get_default_logger()


def results_pipeline(model: QuantumHardwareModel) -> PassManager:
    return (
        PassManager()
        | ResultTransform()
        | IndexMappingAnalysis(model)
        | ErrorMitigation(model)
    )


class LegacyPipelineConfig(PipelineConfig):
    """Configuration for the :class:`LegacyPipeline`, extending :class:`PipelineConfig` with
    configurable connection modes."""

    name: str = "legacy"
    connection_mode: ConnectionMode = ConnectionMode.MANUAL

    @field_validator("connection_mode", mode="before")
    @classmethod
    def _flag_as_string(cls, v):
        if isinstance(v, str):
            return ConnectionMode[v]
        return v


class LegacyExecutePipeline(UpdateablePipeline):
    """A pipeline that executes programs using the legacy backend.

    .. warning::

        This pipeline is for execution purposes only and does not compile programs. Please
        select an appropriate compilation pipeline if you wish to compile programs before
        execution.
    """

    @staticmethod
    def _build_pipeline(
        config: LegacyPipelineConfig,
        model: QuantumHardwareModel,
        target_data: TargetData | None = None,
        engine: None = None,
    ) -> ExecutePipeline:
        if engine is not None:
            log.warning(
                "The engine for the LegacyExecutionPipeline is expected to be provided by "
                "the model, and the provided engine will be ignored."
            )

        if isinstance(model, LiveHardwareModel):
            # Let the Runtime handle startup based on ConnectionMode
            engine = model.create_engine(startup_engine=False)
        else:
            engine = model.create_engine()

        target_data = target_data if target_data is not None else TargetData.default()
        return ExecutePipeline(
            name=config.name,
            model=model,
            target_data=target_data,
            runtime=LegacyRuntime(
                engine=engine,
                results_pipeline=results_pipeline(model),
                connection_mode=config.connection_mode,
            ),
        )



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2024-2025 Oxford Quantum Circuits Ltd

from compiler_config.config import MetricsType
from pydantic import BaseModel, Field, ValidationError, model_validator

snake_cased_flags = [
    val.snake_case_name()
    for val in list(dict(MetricsType.__members__).values())
    if not val.is_composite()
]


class MetricsManager(BaseModel):
    """Stores useful intermediary metrics that are generated during compilation, such as an
    optimised circuit or the instruction count.

    Pydantic version based on :mod:`qat.purr.compiler.metrics` elements.
    """

    enabled_metrics: MetricsType | None = Field(
        default=MetricsType.Experimental, repr=False, exclude=True
    )

    optimized_circuit: str | None = Field(default=None)
    optimized_instruction_count: int | None = Field(default=None)
    physical_qubit_indices: list[int] | None = Field(default=None)

    @model_validator(mode="before")
    def validate_all_fields_exist(cls, value):
        """Validate that all expected MetricsType flags have defined fields."""
        missing_fields = [val for val in snake_cased_flags if val not in cls.model_fields]
        if any(missing_fields):
            raise ValidationError(
                f"MetricsManager is missing type hints for {', '.join(missing_fields)}."
            )
        return value

    def __init__(self, enabled_metrics: MetricsType = MetricsType.Experimental, **kwargs):
        """
        :param enabled_metrics: Which metrics to enable? Default,
            :attr:`MetricsType.Experimental.`"""

        super().__init__(enabled_metrics=enabled_metrics, **kwargs)

    def enable(self, enabled_metrics: MetricsType, overwrite: bool = False):
        """Enable these sets of metrics for collection.

        :param enabled_metrics: Which metrics to enable? Default,
            :attr:`MetricsType.Experimental`.
        :param overwrite: If true, overwrite these metrics with a new collection.
            If false, just enable them in the current collection. Default, true.
        """
        if enabled_metrics is None:
            return

        if overwrite:
            self.enabled_metrics = enabled_metrics
        else:
            self.enabled_metrics = self.enabled_metrics | enabled_metrics

    def enable_metrics(
        self,
        enabled_metrics: MetricsType = MetricsType.Experimental,
        overwrite: bool = True,
    ):
        """Enables the set of metrics in the current collection.

        :param enabled_metrics: Which metrics to enable? Default,
            :attr:`MetricsType.Experimental`.
        :param overwrite: If true, overwrite these metrics with a new collection.
            If false, just enable them in the current collection. Default, true.
        """
        if enabled_metrics is None:
            # log.warning("Attempted to enable metrics with no value. Defaulting.")
            enabled_metrics = MetricsType.Experimental

        # log.info(f"Enabling metrics with flags: {str(enabled_metrics)}")
        self.enable(enabled_metrics, overwrite)

    def are_enabled(self, metric: MetricsType):
        return self.enabled_metrics is not None and metric in self.enabled_metrics

    def record_metric(self, metric: MetricsType, value):
        if not self.are_enabled(metric):
            return

        setattr(self, metric.snake_case_name(), value)

    def get_metric(self, metric: MetricsType):
        return getattr(self, metric.snake_case_name())

    def as_dict(self):
        """Generates a dictionary of the valid metrics."""
        return self.model_dump()

    def merge(self, other: "MetricsManager"):
        if not isinstance(other, type(self)):
            raise TypeError(
                f"Argument other must be of type MetricsManager, not of type: {type(other)}"
            )
        for metric in MetricsType:
            if not metric.is_composite() and self.get_metric(metric) is None:
                self.record_metric(metric, other.get_metric(metric))

        return self



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

import re
from dataclasses import dataclass
from pathlib import Path

from compiler_config.config import Languages

from qat.core.pass_base import AnalysisPass, ResultManager
from qat.core.result_base import ResultInfoMixin

path_regex = re.compile(r"^.+\.(qasm|ll|bc)$")
string_regex = re.compile(
    r"((?P<qasm>OPENQASM) (?P<version>[0-9]+)(?:.[0-9])?;)|(?P<qiskit>@__quantum__qis)"
)


@dataclass
class InputAnalysisResult(ResultInfoMixin):
    language: Languages = Languages.Empty
    raw_input: str | bytes = None


class InputAnalysis(AnalysisPass):
    """Determines the language of the input circuit."""

    def run(self, program: str, res_mgr: ResultManager, *args, **kwargs):
        """
        :param program: A circuit (e.g. QASM or QIR) or the file path for a circuit.
        :param res_mgr: The results manager to save the analysis.
        """

        result = InputAnalysisResult()
        if isinstance(program, bytes) and program.startswith(b"BC"):
            result.language = Languages.QIR
            result.raw_input = program
        elif path_regex.match(program) is not None:
            result.language, result.raw_input = self._process_path_string(program)
        else:
            result.language, result.raw_input = (
                self._process_string(program),
                program,
            )
        res_mgr.add(result)
        if result.language is Languages.Empty:
            raise ValueError("Unable to determine input language.")
        return program

    def _process_path_string(self, path_string):
        path = Path(path_string)
        if path.suffix in (".qasm", ".ll"):
            with path.open() as file:
                string = file.read()
            return self._process_string(string), string
        elif path.suffix == ".bc":
            with path.open("rb") as file:
                bytes_ = file.read()
            return Languages.QIR, bytes_

    def _process_string(self, string):
        match = string_regex.search(string)
        if match is not None:
            if match.group("qasm"):
                version = match.group("version")[0]
                if version == "2":
                    return Languages.Qasm2
                elif version == "3":
                    return Languages.Qasm3
            elif match.group("qiskit"):
                return Languages.QIR
        return Languages.Empty



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from abc import ABC
from numbers import Number

import numpy as np
from pydantic import AfterValidator
from typing_extensions import Annotated

from qat.ir.instructions import Instruction


class QubitInstruction(Instruction, ABC):
    """Denotes that an instruction is done at the level of a qubit (as opposed to a pulse
    channel level instruction).
    """


class GateBase(QubitInstruction, ABC):
    """Base implementation of a gate in a quantum circuit that acts on some qubits, and if
    needed, defined by some parameters.
    """

    def __repr__(self):
        params = tuple(
            [field for field in self.model_dump().values() if isinstance(field, float)]
        )
        return f"{self.__class__.__name__}(qubits={self.qubits}, params={params})"


def constrain_angle(theta: float) -> float:
    """Constrains a rotation angle between [-np.pi, np.pi)."""
    if not isinstance(theta, Number) or np.iscomplex(theta):
        raise ValueError("Angle must be a real number.")
    return (theta + np.pi) % (2 * np.pi) - np.pi


def is_equal_angle(angle1: float, angle2: float) -> bool:
    diff = (angle1 - angle2) % (2 * np.pi)
    return np.isclose(diff, 0.0) or np.isclose(diff, 2 * np.pi)


Angle = Annotated[float, AfterValidator(constrain_angle)]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from collections import defaultdict
from numbers import Number
from typing import Dict, List

from compiler_config.config import MetricsType

from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.devices import PulseChannel
from qat.purr.compiler.instructions import (
    AcquireMode,
    CustomPulse,
    DeviceUpdate,
    Instruction,
    PhaseReset,
    PhaseShift,
    PostProcessing,
    PostProcessType,
    ProcessAxis,
    Pulse,
    Variable,
)
from qat.purr.core.metrics_base import MetricsManager
from qat.purr.core.pass_base import TransformPass
from qat.purr.core.result_base import ResultManager
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


class PhaseOptimisation(TransformPass):
    """
    Extracted from QuantumExecutionEngine.optimize()
    """

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ):
        accum_phaseshifts: Dict[PulseChannel, PhaseShift] = {}
        optimized_instructions: List[Instruction] = []
        for instruction in ir.instructions:
            if isinstance(instruction, PhaseShift) and isinstance(
                instruction.phase, Number
            ):
                if accum_phaseshift := accum_phaseshifts.get(instruction.channel, None):
                    accum_phaseshift.phase += instruction.phase
                else:
                    accum_phaseshifts[instruction.channel] = PhaseShift(
                        instruction.channel, instruction.phase
                    )
            elif isinstance(instruction, (Pulse, CustomPulse)):
                quantum_targets = getattr(instruction, "quantum_targets", [])
                if not isinstance(quantum_targets, List):
                    quantum_targets = [quantum_targets]
                for quantum_target in quantum_targets:
                    if quantum_target in accum_phaseshifts:
                        optimized_instructions.append(accum_phaseshifts.pop(quantum_target))
                optimized_instructions.append(instruction)
            elif isinstance(instruction, PhaseReset):
                for channel in instruction.quantum_targets:
                    accum_phaseshifts.pop(channel, None)
                optimized_instructions.append(instruction)
            else:
                optimized_instructions.append(instruction)

        ir.instructions = optimized_instructions
        met_mgr.record_metric(
            MetricsType.OptimizedInstructionCount, len(optimized_instructions)
        )


class PostProcessingSanitisation(TransformPass):
    """
    Extracted from LiveDeviceEngine.optimize()
    Better pass name/id ?
    """

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ):
        pp_insts = [val for val in ir.instructions if isinstance(val, PostProcessing)]
        discarded = []
        for pp in pp_insts:
            if pp.acquire.mode == AcquireMode.SCOPE:
                if (
                    pp.process == PostProcessType.MEAN
                    and ProcessAxis.SEQUENCE in pp.axes
                    and len(pp.axes) <= 1
                ):
                    discarded.append(pp)

            elif pp.acquire.mode == AcquireMode.INTEGRATOR:
                if (
                    pp.process == PostProcessType.DOWN_CONVERT
                    and ProcessAxis.TIME in pp.axes
                    and len(pp.axes) <= 1
                ):
                    discarded.append(pp)
                if (
                    pp.process == PostProcessType.MEAN
                    and ProcessAxis.TIME in pp.axes
                    and len(pp.axes) <= 1
                ):
                    discarded.append(pp)
        ir.instructions = [val for val in ir.instructions if val not in discarded]
        met_mgr.record_metric(MetricsType.OptimizedInstructionCount, len(ir.instructions))


class DeviceUpdateSanitisation(TransformPass):
    """
    Duplicate DeviceUpdate instructions upsets the device injection mechanism, which causes corruption
    of the HW model.

    In fact, a DeviceInjector is currently 1-1 associated with a DeviceUpdate instruction. When multiple
    DeviceUpdate instructions (sequentially) inject the same "target", the first DeviceInjector assigns the
    (correct) value of the attribute (on the target) to the revert_value. At this point the HW model (or any
    other target kind) is dirty, and any subsequent DeviceInjector updater would surely assign
    the (wrong, usually a placeholder `Variable`) to its revert_value. This results in a corrupt HW model whereby
    reversion wouldn't have the desired effect.

    This pass is a (lazy) fix, which is to analyse when such cases happen and eliminate duplicate DeviceUpdate
    instructions that target THE SAME "attribute" on THE SAME "target" with THE SAME variable.
    """

    def run(
        self,
        ir: InstructionBuilder,
        res_mgr: ResultManager,
        met_mgr: MetricsManager,
        *args,
        **kwargs,
    ):
        target_attr2value = defaultdict(list)
        for inst in ir.instructions:
            if isinstance(inst, DeviceUpdate):
                # target already validated to be a QuantumComponent during initialisation
                if not hasattr(inst.target, inst.attribute):
                    raise ValueError(
                        f"Attempting to assign {inst.value} to non existing attribute {inst.attribute}"
                    )

                if isinstance(inst.value, Variable):
                    target_attr2value[(inst.target, inst.attribute)].append(inst.value)

        for (target, attr), values in target_attr2value.items():
            if len(values) > 1:
                log.warning(
                    f"Multiple DeviceUpdate instructions attempting to update the same attribute '{attr}' on {target}"
                )

            unique_values = set(values)
            if len(unique_values) > 1:
                raise ValueError(
                    f"Cannot update the same attribute '{attr}' on {target} with distinct values {unique_values}"
                )

        new_instructions = []
        for inst in ir.instructions:
            if isinstance(inst, DeviceUpdate) and isinstance(inst.value, Variable):
                if next(iter(target_attr2value[(inst.target, inst.attribute)]), None):
                    target_attr2value[(inst.target, inst.attribute)].clear()
                    new_instructions.append(inst)
            else:
                new_instructions.append(inst)

        ir.instructions = new_instructions



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from compiler_config.config import CompilerConfig

from qat.core.metrics_base import MetricsManager
from qat.core.result_base import ResultManager
from qat.frontend.base import BaseFrontend
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.instructions import DeviceUpdate, Sweep
from qat.purr.utils.logger import get_default_logger

logger = get_default_logger()


class PurrFrontend(BaseFrontend):
    """A frontend for processing purr :class:`InstructionBuilder` source programs.

    The instruction builder will be accepted if it has a model that matches the provided
    model (or if no model is provided), and if it has no sweeps or device injections. These
    are handled by the SweepPipeline.

    :param model: The hardware model to use for compilation.
    """

    def __init__(self, model):
        self.model = model

    def check_and_return_source(self, src):
        """Checks if the source is a valid purr instruction builder.

        :param src: The source program, or path to the program.
        :returns: If the program is determined to not be valid, False is returned.
            Otherwise, the program is returned (and loaded if required).
        """
        if not isinstance(src, InstructionBuilder):
            return False

        if self.model and src.model and self.model != src.model:
            logger.info("InstructionBuilder model does not match the frontend model.")
            return False

        for instr in src.instructions:
            if isinstance(instr, (Sweep, DeviceUpdate)):
                logger.info(
                    "InstructionBuilder contains sweeps or device updates, which are not "
                    "supported by the PurrFrontend. Please use sweep pipelines to handle "
                    "programs with sweeping."
                )
                return False
        return src

    def emit(
        self,
        src: InstructionBuilder,
        res_mgr: ResultManager | None = None,
        met_mgr: MetricsManager | None = None,
        compiler_config: CompilerConfig | None = None,
        **kwargs,
    ) -> InstructionBuilder:
        """Passes through the instruction builder without modifying it."""

        src = self.check_and_return_source(src)
        if src is False:
            raise ValueError("Source is not a valid purr InstructionBuilder.")
        return src



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd

from math import pi

from compiler_config.config import TketOptimizations
from pytket import Bit, Circuit, Qubit
from pytket._tket.architecture import Architecture
from pytket._tket.predicates import MaxNQubitsPredicate
from pytket.circuit import Node
from pytket.passes import SequencePass
from pytket.placement import Placement
from pytket.qasm import circuit_to_qasm_str
from pytket.qasm.qasm import QASMUnsupportedError
from sympy import pi as sympy_pi, sympify

from qat.ir.instruction_builder import InstructionBuilder, QuantumInstructionBuilder
from qat.model.hardware_model import PhysicalHardwareModel as PydHardwareModel
from qat.purr.integrations.tket import (
    TketBuilder as LegacyTketBuilder,
    TketQasmParser,
    _full_stopalize,
    apply_default_transforms,
    check_validity,
    fetch_default_passes,
    optimize_circuit,
)
from qat.purr.utils.logger import get_default_logger
from qat.utils.graphs import get_connected_subgraphs

log = get_default_logger()


def run_tket_optimizations(
    circuit: str | Circuit,
    opts,
    hardware: PydHardwareModel,
    return_as_qasm_str: bool = True,
) -> str | QuantumInstructionBuilder:
    """
    Runs tket-based optimizations and modifications given a Pydantic hardware model.
    Routing will always happen no matter the level.

    Will run optimizations in sections if a full suite fails until a minimal subset of
    passing optimizations is found.
    """

    optimiser = TketOptimisationHelper(circuit, opts, hardware)
    if optimiser.circ is None:
        return circuit
    optimiser.run_optimizations()

    if return_as_qasm_str:
        return optimiser.convert_to_qasm_string()
    else:
        return optimiser.circ


class TketOptimisationHelper:
    """
    Helper class to run pydantic tket optimizations
    """

    def __init__(
        self, circuit: Circuit | str, opts: TketOptimizations, hardware: PydHardwareModel
    ):
        self.hardware = hardware
        self.opts = opts
        if isinstance(circuit, str):
            self.circ = self._get_circuit_from_qasm(circuit)
        else:
            # Helps handle cases where more qubits are allocated than needed
            self.circ = circuit
            self.circ.remove_blank_wires()
        self.coupling_qualities = {}
        self.architecture = None
        self.logical_qubit_map = self._get_logical_qubit_map()

    def run_optimizations(self):
        n_qubits = self.circ.n_qubits
        match n_qubits:
            case 1:
                self.run_one_qubit_optimizations()
            case 2:
                self.run_multi_qubit_optimizations(use_1q_quality=True)
            case _:
                self.run_multi_qubit_optimizations()

    def run_one_qubit_optimizations(self):
        qubit_qualities = {}
        for phys_q, log_q in self.logical_qubit_map.items():
            quality = self.hardware.qubit_quality(phys_q)
            qubit_qualities[log_q] = quality if quality != 1 else 0

        best_qubit = max(qubit_qualities, key=qubit_qualities.get)

        q_map = {Qubit(0): Node(best_qubit), Qubit(best_qubit): Node(0)}
        Placement.place_with_map(self.circ, q_map)

    def run_multi_qubit_optimizations(self, use_1q_quality: bool = False):
        optimisations_failed = False
        if TketOptimizations.DefaultMappingPass not in self.opts:
            couplings = self._get_logical_qubit_couplings()
            optimisations_failed = self._optimise_circuit(couplings)
        else:
            self._calculate_multi_qubit_qualities(use_1q_quality)
            sorted_coupling_qualities = dict(
                sorted(
                    self.coupling_qualities.items(), key=lambda item: item[1], reverse=True
                )
            )

            for quality_level in sorted_coupling_qualities.values():
                filtered_couplings = [
                    key
                    for key, quality in self.coupling_qualities.items()
                    if quality >= quality_level
                ]

                subgraphs_nodes, subgraphs_edges = get_connected_subgraphs(
                    filtered_couplings
                )
                for nodes, edges in zip(subgraphs_nodes, subgraphs_edges):
                    if self.circ.n_qubits <= len(nodes):
                        if not (optimisations_failed := self._optimise_circuit(edges)):
                            break
                    else:
                        optimisations_failed = True
                if not optimisations_failed:
                    break

        if optimisations_failed:
            self._handle_failed_optimisation()

    def convert_to_qasm_string(self):
        try:
            qasm_string = circuit_to_qasm_str(self.circ)
            log.info(f"Number of gates after tket optimization: {self.circ.n_gates}")
            return qasm_string
        except (QASMUnsupportedError, RuntimeError) as e:
            log.warning(
                f"Error generating QASM from Tket circuit: {_full_stopalize(e)}. "
                "Skipping this optimization pass."
            )

    def _get_logical_qubit_couplings(self):
        couplings = []
        for q, coupled_qs in self.hardware.logical_connectivity.items():
            logical_q = self.logical_qubit_map[q]
            for coupled_q in coupled_qs:
                logical_coupled_q = self.logical_qubit_map[coupled_q]
                couplings.append((logical_q, logical_coupled_q))
        return couplings

    def _get_logical_qubit_map(self):
        """
        Without default remapping pass multi-qubit gates don't get moved around, so
        trying to apply them to a limited subset of qubits provides no value.
        """
        return {q_i: i for i, q_i in enumerate(self.hardware.qubits.keys())}

    @staticmethod
    def _get_circuit_from_qasm(qasm_string: str) -> Circuit | None:
        """
        Runs tket-based optimizations and modifications given a Pydantic hardware model.
        Routing will always happen no matter the level.
        """
        try:
            tket_builder: LegacyTketBuilder = TketQasmParser().parse(
                LegacyTketBuilder(), qasm_string
            )
            circ = tket_builder.circuit
            log.info(f"Number of gates before tket optimization: {circ.n_gates}")
            return circ

        except Exception as e:  # Parsing is too fragile, can cause almost any exception.
            log.warning(
                f"Tket failed during QASM parsing with error: {_full_stopalize(e)}. "
                "Skipping this optimization pass."
            )

    def _calculate_multi_qubit_qualities(self, use_1q_quality: bool = False):
        """
        Calculate couplings and coupling qualities from the hardware model.
        """
        for q, coupled_qs in self.hardware.logical_connectivity.items():
            logical_q = self.logical_qubit_map[q]
            for coupled_q in coupled_qs:
                logical_coupled_q = self.logical_qubit_map[coupled_q]

                quality = self.hardware.logical_connectivity_quality[(q, coupled_q)]
                if use_1q_quality:
                    qubit_quality = self.hardware.qubit_quality(q)
                    coupled_qubit_quality = self.hardware.qubit_quality(coupled_q)
                    quality *= qubit_quality * coupled_qubit_quality
                self.coupling_qualities[(logical_q, logical_coupled_q)] = quality

    def _optimise_circuit(self, graph_edges):
        self.architecture = Architecture(graph_edges)
        return not optimize_circuit(self.circ, self.architecture, self.opts)

    def _handle_failed_optimisation(self):
        """
        Check if the optimizations failed.
        If our optimizations failed, but we want the mapping pass, apply that by itself.
        """
        if self.architecture is None:
            raise ValueError(
                "Unable to resolve hardware instance for fall-back optimizations."
            )

        delay_failed = False
        try:
            # DelayMeasure throws on failure, and we want to raise our own errors for
            # this.
            SequencePass(fetch_default_passes(self.architecture, self.opts)).apply(
                self.circ
            )
        except RuntimeError:
            delay_failed = True

        # If the delay fails, try with a more limited subset.
        if delay_failed:
            try:
                # Tket just throws an exception if the list is none, so skip if that's
                # the case.
                default_passes = fetch_default_passes(
                    self.architecture, self.opts, add_delay=False
                )
                if len(default_passes) > 0:
                    SequencePass(default_passes).apply(self.circ)
            except RuntimeError as e:
                message = str(e)
                if MaxNQubitsPredicate.__name__ in message:
                    raise ValueError(
                        f"Circuit uses {len(self.circ.qubits)} qubits, "
                        f"only {len(self.architecture.nodes)} available."
                    ) from e

                raise e

        apply_default_transforms(self.circ, self.architecture, self.opts)
        check_validity(self.circ, self.architecture)


run_pyd_tket_optimizations = run_tket_optimizations


class TketBuilder(InstructionBuilder):
    """Assembles a TKET circuit using the InstructionBuilder interface."""

    def __init__(self, hardware_model: PydHardwareModel):
        super().__init__(hardware_model)
        n_qubits = len(hardware_model.qubits)
        self.circuit = Circuit(n_qubits)
        self._output_variables: dict[int, str] = {}
        self._bit_ctr = 0

    def get_physical_qubit(self, index: int) -> Qubit:
        if index >= len(self.circuit.qubits):
            raise IndexError(
                f"Qubit index {index} out of range for circuit with "
                f"{len(self.circuit.qubits)} qubits."
            )
        return self.circuit.qubits[index]

    def get_logical_qubit(self, index: int) -> Qubit:
        """The TketBuilder doesn't distinguish between physical and logical qubits, so this
        is identical to get_physical_qubit."""
        return self.get_physical_qubit(index)

    @property
    def qubits(self) -> list[Qubit]:
        return self.circuit.qubits

    def X(self, qubit: Qubit, theta: float | None = None):
        if theta is not None:
            self.circuit.Rx(self._normalize_angle(theta), qubit)
        else:
            self.circuit.X(qubit)

    def Y(self, qubit: Qubit, theta: float | None = None):
        if theta is not None:
            self.circuit.Ry(self._normalize_angle(theta), qubit)
        else:
            self.circuit.Y(qubit)
        return self

    def Z(self, qubit: Qubit, theta: float | None = None):
        if theta is not None:
            self.circuit.Rz(self._normalize_angle(theta), qubit)
        else:
            self.circuit.Z(qubit)
        return self

    def U(self, qubit: Qubit, theta: float, phi: float, lamd: float):
        self.circuit.U3(
            self._normalize_angle(theta),
            self._normalize_angle(phi),
            self._normalize_angle(lamd),
            qubit,
        )
        return self

    def swap(self, qubit1: Qubit, qubit2: Qubit):
        self.circuit.SWAP(qubit1, qubit2)
        return self

    def controlled(self, *args):
        raise NotImplementedError("Does not yet support custom control gates.")

    def cX(self, control: Qubit, target: Qubit, theta: float = pi):
        self.circuit.CRx(self._normalize_angle(theta), control, target)
        return self

    def cY(self, control: Qubit, target: Qubit, theta: float = pi):
        self.circuit.CRy(self._normalize_angle(theta), control, target)
        return self

    def cZ(self, control: Qubit, target: Qubit, theta: float = pi):
        self.circuit.CRz(self._normalize_angle(theta), control, target)
        return self

    def cnot(self, control: Qubit, target: Qubit):
        self.circuit.CX(control, target)
        return self

    def ccnot(self, control1: Qubit, control2: Qubit, target: Qubit):
        self.circuit.CCX(control1, control2, target)
        return self

    def ECR(self, qubit1: Qubit, qubit2: Qubit):
        self.circuit.ECR(qubit1, qubit2)
        return self

    def measure_single_shot_z(
        self, qubit: Qubit, output_variable: str | None = None, **kwargs
    ):
        self._output_variables[self._bit_ctr] = output_variable or str(self._bit_ctr)
        cbit = Bit(self._output_variables[self._bit_ctr], self._bit_ctr)
        self.circuit.add_bit(cbit)
        self.circuit.Measure(qubit, cbit)
        self._bit_ctr += 1
        return self

    def reset(self, qubit: Qubit, **kwargs):
        self.circuit.Reset(qubit)
        return self

    def _normalize_angle(self, theta: float) -> float:
        """Tket likes angles to be in units of pi."""
        return theta / pi


class TketToQatIRConverter:
    """Converts a Tket circuit into Qat IR.

    Essentially the "parser" for Tket circuits.

    .. warning::
        This converter is only intended to be used to convert a TKET circuit into QAT IR
        after being parsed from QIR. It does not account for multiple quantum and classical
        registers, and might give undesired behaviour if used outside of this use case.
    """

    @staticmethod
    def convert_parameter(arg: str):
        r"""A parameter stored in a Tket operation is in units of :math:`\pi`. Parameters
        are returned as a string expression, e.g. sometimes containing multiplication and
        division. These expressions are parsed using sympy."""

        return float(sympy_pi * sympify(arg))

    def convert(
        self, qat_builder: InstructionBuilder, tket_builder: TketBuilder
    ) -> InstructionBuilder:
        """Converts a Tket circuit into Qat IR, adding any necesarry assigns and returns.

        :param circuit: Program as a Tket circuit.
        :param result_format: Specifies how measurement results are formatted.
        """

        for command in tket_builder.circuit.to_dict()["commands"]:
            # Retrieves the qubit / clbit indices for each operation
            args = [arg[1][0] for arg in command["args"]]

            match command["op"]["type"]:
                # One-qubit gates
                case "X":
                    qat_builder.X(qat_builder.get_logical_qubit(args[0]))
                case "Y":
                    qat_builder.Y(qat_builder.get_logical_qubit(args[0]))
                case "Z":
                    qat_builder.Z(qat_builder.get_logical_qubit(args[0]))
                case "H":
                    qat_builder.had(qat_builder.get_logical_qubit(args[0]))
                case "SX":
                    qat_builder.SX(qat_builder.get_logical_qubit(args[0]))
                case "SXdg":
                    qat_builder.SXdg(qat_builder.get_logical_qubit(args[0]))
                case "S":
                    qat_builder.S(qat_builder.get_logical_qubit(args[0]))
                case "Sdg":
                    qat_builder.Sdg(qat_builder.get_logical_qubit(args[0]))
                case "T":
                    qat_builder.T(qat_builder.get_logical_qubit(args[0]))
                case "Tdg":
                    qat_builder.Tdg(qat_builder.get_logical_qubit(args[0]))
                case "Rx":
                    qat_builder.X(
                        qat_builder.get_logical_qubit(args[0]),
                        self.convert_parameter(command["op"]["params"][0]),
                    )
                case "Ry":
                    qat_builder.Y(
                        qat_builder.get_logical_qubit(args[0]),
                        self.convert_parameter(command["op"]["params"][0]),
                    )
                case "Rz":
                    qat_builder.Z(
                        qat_builder.get_logical_qubit(args[0]),
                        self.convert_parameter(command["op"]["params"][0]),
                    )
                case "U1":
                    qat_builder.Z(
                        qat_builder.get_logical_qubit(args[0]),
                        self.convert_parameter(command["op"]["params"][0]),
                    )
                case "U2":
                    qat_builder.U(
                        qat_builder.get_logical_qubit(args[0]),
                        pi / 2,
                        self.convert_parameter(command["op"]["params"][0]),
                        self.convert_parameter(command["op"]["params"][1]),
                    )
                case "U3":
                    qat_builder.U(
                        qat_builder.get_logical_qubit(args[0]),
                        self.convert_parameter(command["op"]["params"][0]),
                        self.convert_parameter(command["op"]["params"][1]),
                        self.convert_parameter(command["op"]["params"][2]),
                    )

                # Two-qubit gates
                case "CX":
                    qat_builder.cnot(
                        qat_builder.get_logical_qubit(args[0]),
                        qat_builder.get_logical_qubit(args[1]),
                    )
                case "ECR":
                    qat_builder.ECR(
                        qat_builder.get_logical_qubit(args[0]),
                        qat_builder.get_logical_qubit(args[1]),
                    )
                case "SWAP":
                    qat_builder.swap(
                        qat_builder.get_logical_qubit(args[0]),
                        qat_builder.get_logical_qubit(args[1]),
                    )

                # Operations
                case "Measure":
                    output_var = tket_builder._output_variables[int(str(args[1]))]
                    qat_builder.measure_single_shot_z(
                        qat_builder.get_logical_qubit(args[0]),
                        output_variable=output_var,
                    )
                case "Barrier":
                    qat_builder.synchronize(
                        [qat_builder.get_logical_qubit(arg) for arg in args]
                    )
                case "Reset":
                    qat_builder.reset(qat_builder.get_logical_qubit(args[0]))
                case _:
                    raise NotImplementedError(
                        f"Command {command['op']['type']} not implemented."
                    )

        qat_builder += tket_builder
        return qat_builder



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd
import abc
import os
import tempfile
from typing import Tuple

import regex
from compiler_config.config import (
    CompilerConfig,
    Languages,
    Tket,
    TketOptimizations,
    get_optimizer_config,
)

from qat.purr.backends.calibrations.remote import find_calibration
from qat.purr.backends.realtime_chip_simulator import get_default_RTCS_hardware
from qat.purr.compiler.builders import InstructionBuilder
from qat.purr.compiler.metrics import CompilationMetrics
from qat.purr.compiler.optimisers import DefaultOptimizers
from qat.purr.compiler.runtime import execute_instructions, get_builder, get_model
from qat.purr.integrations.qasm import get_qasm_parser
from qat.purr.integrations.qir import QIRParser
from qat.purr.utils.logger import get_default_logger
from qat.purr.utils.logging_utils import log_duration

log = get_default_logger()


def _get_file_contents(file_path):
    """Get Program contents from a file."""
    with open(file_path) as ifile:
        return ifile.read()


path_regex = regex.compile(r"^.+\.(qasm|ll|bc)$")


class LanguageFrontend(abc.ABC):
    def _build_instructions(
        self,
        quantum_builder: InstructionBuilder,
        hardware,
        compiler_config: CompilerConfig,
    ):
        instructions = (
            get_builder(hardware)
            .repeat(
                compiler_config.repeats,
                compiler_config.repetition_period,
                compiler_config.passive_reset_time,
            )
            .add(quantum_builder)
        )
        return instructions

    def _execute(
        self,
        hardware,
        compiler_config: CompilerConfig,
        instructions,
        *args,
        **kwargs,
    ):
        calibrations = [
            find_calibration(arg) for arg in compiler_config.active_calibrations
        ]

        return execute_instructions(
            hardware,
            instructions,
            compiler_config,
            calibrations,
        )

    def _default_common_args(self, hardware=None, compiler_config=None):
        hardware = hardware or get_default_RTCS_hardware()
        compiler_config = compiler_config or CompilerConfig()
        return hardware, compiler_config

    @abc.abstractmethod
    def parse(
        self, program_str: str, hardware, compiler_config: CompilerConfig
    ) -> Tuple[InstructionBuilder, CompilationMetrics]: ...

    @abc.abstractmethod
    def execute(
        self,
        instructions: InstructionBuilder,
        hardware,
        compiler_config: CompilerConfig,
        *args,
        **kwargs,
    ): ...


class QIRFrontend(LanguageFrontend):
    def _parse_from_file(
        self, path_or_str: str, hardware=None, compiler_config: CompilerConfig = None
    ):
        hardware, compiler_config = self._default_common_args(hardware, compiler_config)

        metrics = CompilationMetrics()
        metrics.enable(compiler_config.metrics)

        if compiler_config.optimizations is None:
            compiler_config.optimizations = get_optimizer_config(Languages.QIR)

        with log_duration("Compilation completed, took {} seconds."):
            log.info(f"Processing QIR  as parser and {str(hardware)} as hardware.")

            optimizations = compiler_config.optimizations
            if (
                isinstance(optimizations, Tket)
                and optimizations.tket_optimizations != TketOptimizations.Empty
            ):
                quantum_builder = DefaultOptimizers(metrics).optimize_qir(
                    path_or_str,
                    get_model(hardware),
                    compiler_config.optimizations,
                    compiler_config.results_format.format,
                )
            else:
                parser = QIRParser(
                    hardware,
                    builder=get_model(hardware).create_builder(),
                    results_format=compiler_config.results_format.format,
                )
                quantum_builder = parser.parse(path_or_str)

            return (
                self._build_instructions(quantum_builder, hardware, compiler_config),
                metrics,
            )

    def parse(
        self, path_or_str: str, hardware=None, compiler_config: CompilerConfig = None
    ):
        # Parse from file
        if not os.path.exists(path_or_str):
            suffix = ".bc" if isinstance(path_or_str, bytes) else ".ll"
            with tempfile.NamedTemporaryFile(suffix=suffix, delete=False) as fp:
                if suffix == ".ll":
                    fp.write(path_or_str.encode())
                else:
                    fp.write(path_or_str)
                fp.close()
                try:
                    return self._parse_from_file(fp.name, hardware, compiler_config)
                finally:
                    os.remove(fp.name)
        return self._parse_from_file(path_or_str, hardware, compiler_config)

    def execute(
        self,
        instructions: InstructionBuilder,
        hardware=None,
        compiler_config: CompilerConfig = None,
        *args,
        **kwargs,
    ):
        hardware, compiler_config = self._default_common_args(hardware, compiler_config)

        with log_duration("Execution completed, took {} seconds."):
            return self._execute(hardware, compiler_config, instructions, *args, **kwargs)

    def parse_and_execute(
        self,
        qir_file: str,
        hardware=None,
        compiler_config: CompilerConfig = None,
        *args,
        **kwargs,
    ):
        instructions, parse_metrics = self.parse(qir_file, hardware, compiler_config)
        result, execution_metrics = self.execute(
            instructions, hardware, compiler_config, *args, **kwargs
        )
        execution_metrics.merge(parse_metrics)
        return result, execution_metrics


class QASMFrontend(LanguageFrontend):
    def parse(
        self, path_or_str: str, hardware=None, compiler_config: CompilerConfig = None
    ):
        # Parse from contents
        qasm_string = path_or_str
        if os.path.isfile(path_or_str):
            qasm_string = _get_file_contents(path_or_str)

        hardware, compiler_config = self._default_common_args(hardware, compiler_config)

        metrics = CompilationMetrics()
        metrics.enable(compiler_config.metrics)

        parser = get_qasm_parser(qasm_string)
        if compiler_config.optimizations is None:
            compiler_config.optimizations = get_optimizer_config(parser.parser_language())

        with log_duration("Compilation completed, took {} seconds."):
            log.info(
                f"Processing QASM with {str(parser)} as parser and {str(hardware)} "
                "as hardware."
            )

            qasm_string = DefaultOptimizers(metrics).optimize_qasm(
                qasm_string, get_model(hardware), compiler_config.optimizations
            )

            if compiler_config.results_format.format is not None:
                parser.results_format = compiler_config.results_format.format

            quantum_builder = parser.parse(get_builder(hardware), qasm_string)
            return (
                self._build_instructions(quantum_builder, hardware, compiler_config),
                metrics,
            )

    def execute(
        self,
        instructions: InstructionBuilder,
        hardware=None,
        compiler_config: CompilerConfig = None,
        *args,
        **kwargs,
    ):
        hardware, compiler_config = self._default_common_args(hardware, compiler_config)

        with log_duration("Execution completed, took {} seconds."):
            return self._execute(hardware, compiler_config, instructions, *args, **kwargs)

    def parse_and_execute(
        self,
        qasm_string: str,
        hardware=None,
        compiler_config: CompilerConfig = None,
        *args,
        **kwargs,
    ):
        """
        Execute a qasm string against a particular piece of hardware. Initializes a
        default qubit simulator if no hardware provided.
        """
        instructions, parse_metrics = self.parse(qasm_string, hardware, compiler_config)
        result, execution_metrics = self.execute(
            instructions, hardware, compiler_config, *args, **kwargs
        )
        parse_metrics.merge(execution_metrics)
        return result, parse_metrics



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023-2025 Oxford Quantum Circuits Ltd
import re
from typing import List, Optional, Tuple, Union
from warnings import warn

import numpy as np
from compiler_config.config import (
    ErrorMitigationConfig,
    QuantumResultsFormat,
    ResultsFormatting,
)
from qiskit import QiskitError, QuantumCircuit, transpile
from qiskit.transpiler import CouplingMap
from qiskit.transpiler.passes import CheckGateDirection, CheckMap
from qiskit_aer import AerSimulator
from qiskit_aer.backends.backendconfiguration import AerBackendConfiguration

from qat.core.config.configure import get_config
from qat.purr.backends.echo import (
    Connectivity,
    add_direction_couplings_to_hardware,
    apply_setup_to_hardware,
    generate_connectivity,
)
from qat.purr.compiler.builders import Axis, InstructionBuilder
from qat.purr.compiler.devices import PulseChannel, Qubit
from qat.purr.compiler.error_mitigation.readout_mitigation import get_readout_mitigation
from qat.purr.compiler.execution import InstructionExecutionEngine
from qat.purr.compiler.hardware_models import QuantumHardwareModel
from qat.purr.compiler.instructions import (
    Assign,
    Instruction,
    Return,
    Variable,
    is_generated_name,
)
from qat.purr.compiler.runtime import QuantumRuntime
from qat.purr.utils.logger import get_default_logger

log = get_default_logger()


def get_default_qiskit_hardware(
    qubit_count=20,
    noise_model=None,
    strict_placement=True,
    connectivity: Optional[Union[Connectivity, List[Tuple[int, int]]]] = None,
) -> "QiskitHardwareModel":
    """
    Creates a hardware model compatible with the Qiskit simulator.

    If `strict_placement=True`, circuits can only be executed when circuit
    intructions act on adjacent qubits in the coupling map.
    """
    model = QiskitHardwareModel(qubit_count, noise_model)
    model.strict_placement = strict_placement

    if isinstance(connectivity, Connectivity):
        connectivity = generate_connectivity(connectivity, qubit_count)
    connectivity = connectivity or generate_connectivity(Connectivity.Ring, qubit_count)

    model = apply_setup_to_hardware(model, qubit_count, connectivity)
    return add_direction_couplings_to_hardware(model, connectivity)


class QiskitBuilder(InstructionBuilder):
    """Builder around QASM circuits."""

    def __init__(
        self,
        hardware_model: QuantumHardwareModel,
        qubit_count: int,
        instructions: InstructionBuilder = None,
    ):
        super().__init__(hardware_model=hardware_model, instructions=instructions)
        self.circuit = QuantumCircuit(qubit_count, qubit_count)
        self.shot_count = hardware_model.default_repeat_count
        self.bit_count = 0
        self.bit_ordering = {}

    def measure(self, target: Qubit, *args, **kwargs) -> "InstructionBuilder":
        self.circuit.measure(target.index, self.bit_count)
        # keep track of the ordering of measurements
        if len(args) >= 2:
            self.bit_ordering[args[1]] = self.bit_count
        self.bit_count = self.bit_count + 1
        return self

    def R(self, axis: Axis, target: Union[Qubit, PulseChannel], radii=None):
        qb, _ = self.model.resolve_qb_pulse_channel(target)
        if radii is None:
            radii = np.pi

        if axis == Axis.X:
            self.circuit.rx(radii, qb.index)
        elif axis == Axis.Y:
            self.circuit.ry(radii, qb.index)
        elif axis == Axis.Z:
            self.circuit.rz(radii, qb.index)

        return self

    def Z(self, target: Union[Qubit, PulseChannel], radii=np.pi):
        qb, _ = self.model.resolve_qb_pulse_channel(target)
        self.circuit.rz(radii, qb.index)
        return self

    def X(self, target: Union[Qubit, PulseChannel], radii=np.pi):
        qb, _ = self.model.resolve_qb_pulse_channel(target)
        self.circuit.rx(radii, qb.index)
        return self

    def Y(self, target: Union[Qubit, PulseChannel], radii=np.pi):
        qb, _ = self.model.resolve_qb_pulse_channel(target)
        self.circuit.ry(radii, qb.index)
        return self

    def swap(self, target: Qubit, destination: Qubit):
        self.circuit.swap(target.index, destination.index)
        return self

    def delay(self, target: Union[Qubit, PulseChannel], time: float):
        qb, _ = self.model.resolve_qb_pulse_channel(target)
        self.circuit.delay(time, qb.index)
        return self

    def cR(
        self,
        axis: Axis,
        controllers: Union[Qubit, List[Qubit]],
        target: Qubit,
        theta: float,
    ):
        if len(controllers) > 1:
            raise ValueError(
                "Cannot perform generic controlled operation across multiple qubits."
            )

        controllers = controllers[0]

        if axis == Axis.X:
            self.circuit.crx(theta, controllers, target.index)
        elif axis == Axis.Y:
            self.circuit.cry(theta, controllers, target.index)
        elif axis == Axis.Z:
            self.circuit.crz(theta, controllers, target.index)

        return self

    def cnot(self, control: Qubit, target_qubit: Qubit):
        self.circuit.cx(control.index, target_qubit.index)
        return self

    def ccnot(self, cone: Qubit, ctwo: Qubit, target_qubit: Qubit):
        self.circuit.ccx(cone.index, ctwo.index, target_qubit.index)
        return self

    def ECR(self, control: Qubit, target: Qubit):
        self.circuit.ecr(control.index, target.index)
        return self

    def had(self, qubit: Qubit):
        self.circuit.h(qubit.index)
        return self

    def reset(self, qubits: Union[Qubit, List[Qubit]]):
        if not isinstance(qubits, list):
            qubits = [qubits]

        for qb in qubits:
            self.circuit.reset(qb.index)

        return self

    def repeat(self, repeat_value: int = None, *args, **kwargs):
        # Compiler config may attempt to pass None
        if repeat_value is not None:
            self.shot_count = repeat_value
        else:
            self.shot_count = self.model.default_repeat_count
        return self

    def clear(self):
        self.circuit.clear()
        self.bit_count = 0
        self.bit_ordering = {}

    def merge_builder(self, other_builder: InstructionBuilder):
        """
        overloaded merge_builder Ensure that circuits are copied too
        As well as the rest of the base merging operations.
        """
        self.circuit.compose(other_builder.circuit, inplace=True)
        self.bit_count = other_builder.bit_count
        self.bit_ordering = other_builder.bit_ordering

        return super().merge_builder(other_builder)


class QiskitHardwareModel(QuantumHardwareModel):
    def __init__(self, qubit_count, noise_model=None, strict_placement=True):
        self.qubit_count = qubit_count
        self.noise_model = noise_model
        self.strict_placement = strict_placement
        super().__init__()

    def create_runtime(self, existing_engine: InstructionExecutionEngine = None):
        # We aren't going to use the engine anyway.
        return QiskitRuntime(existing_engine or self.create_engine())

    def create_builder(self) -> InstructionBuilder:
        return QiskitBuilder(hardware_model=self, qubit_count=self.qubit_count)

    def create_engine(self) -> InstructionExecutionEngine:
        return QiskitEngine(hardware_model=self)


def verify_placement(coupling_map, circuit):
    """
    Check that the circuit can be directly mapped according to the
    wire map defined by the coupling map

    Raises if placement cannot happen without swaps
    """

    cmap = CouplingMap(coupling_map)
    checker = CheckMap(cmap)
    checker(circuit)
    if not checker.property_set["is_swap_mapped"]:
        raise RuntimeError(
            f"Cannot achieve placement on set couplings with operation:"
            f" {checker.property_set['check_map_msg']}"
        )

    checker = CheckGateDirection(cmap)
    checker(circuit)
    if not checker.property_set["is_direction_mapped"]:
        warn(
            "Placement would fail on a real QPU with the same topology due to invalid gate "
            "directions. This is currently allowed for the Qiskit simulator, but will be "
            "changed in future versions of QAT.",
            DeprecationWarning,
            stacklevel=2,
        )


class QiskitEngine(InstructionExecutionEngine):
    def __init__(self, hardware_model: QiskitHardwareModel = None):
        super().__init__(hardware_model)

    def run_calibrations(self, qubits_to_calibrate=None):
        pass

    def _run_simulator(self, builder, aer_config, method, coupling_map):
        qasm_sim = AerSimulator(
            aer_config,
            noise_model=builder.model.noise_model,
            method=method,
            **get_config().SIMULATION.QISKIT.OPTIONS,
        )
        initial_layout = {qb: i for i, qb in enumerate(builder.circuit.qubits)}
        job = qasm_sim.run(
            transpile(
                builder.circuit,
                initial_layout=initial_layout,
                target=qasm_sim.target,
            ),
            shots=builder.shot_count,
        )
        return job.result()

    def execute(self, builder: QiskitBuilder):
        """
        Execute a circuit using Qiskit's AerSimulator as a backend. Options for the
        simulator can be provided using qatconfig.SIMULATION.QISKIT.

        For more information, see:
        https://docs.quantum.ibm.com/api/qiskit/0.37/qiskit.providers.aer.AerSimulator
        """
        if not isinstance(builder, QiskitBuilder):
            raise ValueError(
                "Contravariance is not possible with QiskitEngine. "
                "execute input must be a QiskitBuilder"
            )

        coupling_map = None
        if any(self.model.qubit_direction_couplings):
            coupling_map = [
                list(coupling.direction)
                for coupling in self.model.qubit_direction_couplings
            ]

        circuit = builder.circuit
        if self.model.strict_placement:
            verify_placement(coupling_map, circuit)

        # With no coupling map the backend defaults to create couplings for qubit count, which
        # defaults to 30. So we change that.
        aer_config = AerBackendConfiguration.from_dict(
            AerSimulator._DEFAULT_CONFIGURATION
            | {"open_pulse": False, "coupling_map": coupling_map}
        )
        aer_config.n_qubits = self.model.qubit_count

        # AerSimulator can call a range of backends. We allow these to be choosen, along with
        # any relevant arguments.

        qiskitconfig = get_config().SIMULATION.QISKIT
        try:
            # Determine the sequence of backends to try
            method = qiskitconfig.METHOD
            seq = qiskitconfig.FALLBACK_SEQUENCE
            if method in seq:
                seq.remove(method)
            seq.insert(0, method)

            # Run the simulator
            results = {}
            for method in seq:
                results = self._run_simulator(
                    builder,
                    aer_config,
                    method,
                    coupling_map,
                )

                if results.results[0].success:
                    break
                log.warning(f"AerSimulator simulation with method {method} failed.")
            distribution = results.get_counts(circuit)
        except QiskitError as e:
            raise ValueError(f"QiskitError while running Qiskit circuit: {str(e)}")

        # return the measurements in the correct order
        assigns = {}
        returns = []
        for inst in builder.instructions:
            if isinstance(inst, Assign):
                assigns[inst.name] = inst.value
            elif isinstance(inst, Return):
                returns.extend(inst.variables)
        # trim the qiskit returns to the number of measurements and revers the index order
        trimmed = {
            key[-builder.bit_count :][::-1]: value for key, value in distribution.items()
        }
        if len(returns) > 0:
            task_results = {}
            for creg in returns:
                creg_result = {}
                c_indices = [
                    (builder.bit_ordering[ref.name] if isinstance(ref, Variable) else None)
                    for ref in assigns[creg]
                ]
                for key, value in trimmed.items():
                    key_list = list(key)
                    new_key = "".join(
                        [
                            key_list[ind] if isinstance(ind, int) else "0"
                            for ind in c_indices
                        ]
                    )
                    if new_key in creg_result:
                        creg_result[new_key] += value
                    else:
                        creg_result[new_key] = value
                task_results[creg] = creg_result
        else:
            task_results = trimmed

        return (
            (task_results, results.results[0].metadata)
            if qiskitconfig.ENABLE_METADATA
            else task_results
        )

    def optimize(self, instructions: List[Instruction]) -> List[Instruction]:
        log.info("No optimize implemented for QiskitEngine")
        return instructions

    def validate(self, instructions: List[Instruction]):
        pass


cl2qu_index_pattern = re.compile(r"(.*)\[(?P<clbit_index>[0-9]+)\]_(?P<qubit_index>[0-9]+)")


def get_cl2qu_index_mapping(instructions):
    mapping = {}
    for instruction in instructions:
        if not isinstance(instruction, Assign):
            continue
        for value in instruction.value:
            result = cl2qu_index_pattern.match(value.name)
            if result is None:
                raise ValueError(
                    f"Could not extract cl register index from {instruction.output_variable}"
                )
            mapping[result.group("clbit_index")] = int(result.group("qubit_index"))
    return mapping


class QiskitRuntime(QuantumRuntime):
    def _apply_error_mitigation(self, results, instructions, error_mitigation):
        if error_mitigation is None:
            return results

        # TODO: add support for multiple registers
        # TODO: reconsider results length
        if len(results) > 1:
            raise ValueError(
                "Cannot have multiple registers in conjunction with readout error mitigation."
            )

        mapping = get_cl2qu_index_mapping(instructions)

        for mitigator in get_readout_mitigation(error_mitigation):
            new_result = mitigator.apply_error_mitigation(results, mapping, self.model)
            results[mitigator.name] = new_result

        return results  # TODO: new results object

    def execute(
        self,
        builder: QiskitBuilder,
        results_format: ResultsFormatting = None,
        repeats: int = None,
        error_mitigation: ErrorMitigationConfig = None,
    ):
        """
        Execute a circuit using Qiskit's AerSimulator as a backend. Options for the
        simulator can be provided using qatconfig.SIMULATION.QISKIT.

        For more information, see:
        https://docs.quantum.ibm.com/api/qiskit/0.37/qiskit.providers.aer.AerSimulator
        """
        if not isinstance(builder, QiskitBuilder):
            raise ValueError("Wrong builder type passed to QASM runtime.")
        results = self.engine.execute(builder)

        qiskitconfig = get_config().SIMULATION.QISKIT
        if qiskitconfig.ENABLE_METADATA:
            (results, metadata) = results

        format_flags = (
            results_format.transforms
            if isinstance(results_format, QuantumResultsFormat)
            else results_format
        )
        if format_flags is None or ResultsFormatting.BinaryCount not in format_flags:
            log.warning(
                "The results formatting `BinaryCount` was not found in the formatting "
                "flags. Please note that the Qiskit runtime only currently supports "
                "results returned as a binary count."
            )

        results = self._apply_error_mitigation(
            results, builder.instructions, error_mitigation
        )

        if all([is_generated_name(k) for k in results.keys()]):
            if len(results) == 1:
                results = list(results.values())[0]
            else:
                squashed_results = list(results.values())
                if all(isinstance(val, np.ndarray) for val in squashed_results):
                    results = np.array(squashed_results)
                else:
                    results = squashed_results
        return (results, metadata) if qiskitconfig.ENABLE_METADATA else results


class QiskitBuilderWrapper:
    """A wrapped around the Qiskit builder to smoothly integrate into the pipeline API.

    A mismatch in the API between the :class:`QuantumExecutionEngine` and
    :class:`QiskitEngine` means that we can't easily implement Qiskit into pipelines: this
    is a little "hack" to make the API match.
    """

    def __init__(self, builder: QiskitBuilder):
        self.builder = builder

    @property
    def instructions(self):
        return self.builder

    @property
    def calibration_id(self):
        return ""



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd

from qat.model.loaders.purr import EchoModelLoader
from qat.pipelines.updateable import PipelineConfig

from .compile import WaveformV1CompilePipeline
from .execute import EchoExecutePipeline
from .full import EchoPipeline


def _create_pipeline_instance(num_qubits: int) -> EchoPipeline:
    return EchoPipeline(
        config=PipelineConfig(name=f"echo{num_qubits}"),
        loader=EchoModelLoader(qubit_count=num_qubits),
    ).pipeline


echo8 = _create_pipeline_instance(8)
echo16 = _create_pipeline_instance(16)
echo32 = _create_pipeline_instance(32)


__all__ = [
    "EchoPipeline",
    "EchoExecutePipeline",
    "echo8",
    "echo16",
    "echo32",
    "PipelineConfig",
    "WaveformV1CompilePipeline",
]



# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2025 Oxford Quantum Circuits Ltd
from qat.engines.native import (
    ConnectionMixin as ConnectionMixin,
    NativeEngine as NativeEngine,
)
from qat.engines.zero import ZeroEngine as ZeroEngine

DefaultEngine = ZeroEngine
